# Copyright 2019 The gRPC Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests behavior of the Call classes."""

import asyncio
import datetime
import logging
import random
import unittest

import grpc
from grpc.experimental import aio

from src.proto.grpc.testing import messages_pb2
from src.proto.grpc.testing import test_pb2_grpc
from tests_aio.unit._constants import UNREACHABLE_TARGET
from tests_aio.unit._test_base import AioTestBase
from tests_aio.unit._test_server import start_test_server

_SHORT_TIMEOUT_S = datetime.timedelta(seconds=1).total_seconds()

_NUM_STREAM_RESPONSES = 5
_RESPONSE_PAYLOAD_SIZE = 42
_REQUEST_PAYLOAD_SIZE = 7
_LOCAL_CANCEL_DETAILS_EXPECTATION = "Locally cancelled by application!"
_RESPONSE_INTERVAL_US = int(_SHORT_TIMEOUT_S * 1000 * 1000)
_INFINITE_INTERVAL_US = 2**31 - 1

_NONDETERMINISTIC_ITERATIONS = 50
_NONDETERMINISTIC_SERVER_SLEEP_MAX_US = 1000


class _MulticallableTestMixin:
    async def setUp(self):
        address, self._server = await start_test_server()
        self._channel = aio.insecure_channel(address)
        self._stub = test_pb2_grpc.TestServiceStub(self._channel)

    async def tearDown(self):
        await self._channel.close()
        await self._server.stop(None)


class TestUnaryUnaryCall(_MulticallableTestMixin, AioTestBase):
    async def test_call_to_string(self):
        call = self._stub.UnaryCall(messages_pb2.SimpleRequest())

        self.assertTrue(str(call) is not None)
        self.assertTrue(repr(call) is not None)

        await call

        self.assertTrue(str(call) is not None)
        self.assertTrue(repr(call) is not None)

    async def test_call_ok(self):
        call = self._stub.UnaryCall(messages_pb2.SimpleRequest())

        self.assertFalse(call.done())

        response = await call

        self.assertTrue(call.done())
        self.assertIsInstance(response, messages_pb2.SimpleResponse)
        self.assertEqual(await call.code(), grpc.StatusCode.OK)

        # Response is cached at call object level, reentrance
        # returns again the same response
        response_retry = await call
        self.assertIs(response, response_retry)

    async def test_call_rpc_error(self):
        async with aio.insecure_channel(UNREACHABLE_TARGET) as channel:
            stub = test_pb2_grpc.TestServiceStub(channel)

            call = stub.UnaryCall(messages_pb2.SimpleRequest())

            with self.assertRaises(aio.AioRpcError) as exception_context:
                await call

            self.assertEqual(
                grpc.StatusCode.UNAVAILABLE, exception_context.exception.code()
            )

            self.assertTrue(call.done())
            self.assertEqual(grpc.StatusCode.UNAVAILABLE, await call.code())

    async def test_call_code_awaitable(self):
        call = self._stub.UnaryCall(messages_pb2.SimpleRequest())
        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_call_details_awaitable(self):
        call = self._stub.UnaryCall(messages_pb2.SimpleRequest())
        self.assertEqual("", await call.details())

    async def test_call_initial_metadata_awaitable(self):
        call = self._stub.UnaryCall(messages_pb2.SimpleRequest())
        self.assertEqual(aio.Metadata(), await call.initial_metadata())

    async def test_call_trailing_metadata_awaitable(self):
        call = self._stub.UnaryCall(messages_pb2.SimpleRequest())
        self.assertEqual(aio.Metadata(), await call.trailing_metadata())

    async def test_call_initial_metadata_cancelable(self):
        coro_started = asyncio.Event()
        call = self._stub.UnaryCall(messages_pb2.SimpleRequest())

        async def coro():
            coro_started.set()
            await call.initial_metadata()

        task = self.loop.create_task(coro())
        await coro_started.wait()
        task.cancel()

        # Test that initial metadata can still be asked thought
        # a cancellation happened with the previous task
        self.assertEqual(aio.Metadata(), await call.initial_metadata())

    async def test_call_initial_metadata_multiple_waiters(self):
        call = self._stub.UnaryCall(messages_pb2.SimpleRequest())

        async def coro():
            return await call.initial_metadata()

        task1 = self.loop.create_task(coro())
        task2 = self.loop.create_task(coro())

        await call
        expected = [aio.Metadata() for _ in range(2)]
        self.assertEqual(expected, await asyncio.gather(*[task1, task2]))

    async def test_call_code_cancelable(self):
        coro_started = asyncio.Event()
        call = self._stub.UnaryCall(messages_pb2.SimpleRequest())

        async def coro():
            coro_started.set()
            await call.code()

        task = self.loop.create_task(coro())
        await coro_started.wait()
        task.cancel()

        # Test that code can still be asked thought
        # a cancellation happened with the previous task
        self.assertEqual(grpc.StatusCode.OK, await call.code())

    async def test_call_code_multiple_waiters(self):
        call = self._stub.UnaryCall(messages_pb2.SimpleRequest())

        async def coro():
            return await call.code()

        task1 = self.loop.create_task(coro())
        task2 = self.loop.create_task(coro())

        await call

        self.assertEqual(
            [grpc.StatusCode.OK, grpc.StatusCode.OK],
            await asyncio.gather(task1, task2),
        )

    async def test_cancel_unary_unary(self):
        call = self._stub.UnaryCall(messages_pb2.SimpleRequest())

        self.assertFalse(call.cancelled())

        self.assertTrue(call.cancel())
        self.assertFalse(call.cancel())

        with self.assertRaises(asyncio.CancelledError):
            await call

        # The info in the RpcError should match the info in Call object.
        self.assertTrue(call.cancelled())
        self.assertEqual(await call.code(), grpc.StatusCode.CANCELLED)
        self.assertEqual(
            await call.details(), "Locally cancelled by application!"
        )

    async def test_cancel_unary_unary_in_task(self):
        coro_started = asyncio.Event()
        call = self._stub.EmptyCall(messages_pb2.SimpleRequest())

        async def another_coro():
            coro_started.set()
            await call

        task = self.loop.create_task(another_coro())
        await coro_started.wait()

        self.assertFalse(task.done())
        task.cancel()

        self.assertEqual(grpc.StatusCode.CANCELLED, await call.code())

        with self.assertRaises(asyncio.CancelledError):
            await task

    async def test_passing_credentials_fails_over_insecure_channel(self):
        call_credentials = grpc.composite_call_credentials(
            grpc.access_token_call_credentials("abc"),
            grpc.access_token_call_credentials("def"),
        )
        with self.assertRaisesRegex(
            aio.UsageError, "Call credentials are only valid on secure channels"
        ):
            self._stub.UnaryCall(
                messages_pb2.SimpleRequest(), credentials=call_credentials
            )


class TestUnaryStreamCall(_MulticallableTestMixin, AioTestBase):
    async def test_call_rpc_error(self):
        channel = aio.insecure_channel(UNREACHABLE_TARGET)
        request = messages_pb2.StreamingOutputCallRequest()
        stub = test_pb2_grpc.TestServiceStub(channel)
        call = stub.StreamingOutputCall(request)

        with self.assertRaises(aio.AioRpcError) as exception_context:
            async for response in call:
                pass

        self.assertEqual(
            grpc.StatusCode.UNAVAILABLE, exception_context.exception.code()
        )

        self.assertTrue(call.done())
        self.assertEqual(grpc.StatusCode.UNAVAILABLE, await call.code())
        await channel.close()

    async def test_cancel_unary_stream(self):
        # Prepares the request
        request = messages_pb2.StreamingOutputCallRequest()
        for _ in range(_NUM_STREAM_RESPONSES):
            request.response_parameters.append(
                messages_pb2.ResponseParameters(
                    size=_RESPONSE_PAYLOAD_SIZE,
                    interval_us=_RESPONSE_INTERVAL_US,
                )
            )

        # Invokes the actual RPC
        call = self._stub.StreamingOutputCall(request)
        self.assertFalse(call.cancelled())

        response = await call.read()
        self.assertIs(type(response), messages_pb2.StreamingOutputCallResponse)
        self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))

        self.assertTrue(call.cancel())
        self.assertEqual(grpc.StatusCode.CANCELLED, await call.code())
        self.assertEqual(
            _LOCAL_CANCEL_DETAILS_EXPECTATION, await call.details()
        )
        self.assertFalse(call.cancel())

        with self.assertRaises(asyncio.CancelledError):
            await call.read()
        self.assertTrue(call.cancelled())

    async def test_multiple_cancel_unary_stream(self):
        # Prepares the request
        request = messages_pb2.StreamingOutputCallRequest()
        for _ in range(_NUM_STREAM_RESPONSES):
            request.response_parameters.append(
                messages_pb2.ResponseParameters(
                    size=_RESPONSE_PAYLOAD_SIZE,
                    interval_us=_RESPONSE_INTERVAL_US,
                )
            )

        # Invokes the actual RPC
        call = self._stub.StreamingOutputCall(request)
        self.assertFalse(call.cancelled())

        response = await call.read()
        self.assertIs(type(response), messages_pb2.StreamingOutputCallResponse)
        self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))

        self.assertTrue(call.cancel())
        self.assertFalse(call.cancel())
        self.assertFalse(call.cancel())
        self.assertFalse(call.cancel())

        with self.assertRaises(asyncio.CancelledError):
            await call.read()

    async def test_early_cancel_unary_stream(self):
        """Test cancellation before receiving messages."""
        # Prepares the request
        request = messages_pb2.StreamingOutputCallRequest()
        for _ in range(_NUM_STREAM_RESPONSES):
            request.response_parameters.append(
                messages_pb2.ResponseParameters(
                    size=_RESPONSE_PAYLOAD_SIZE,
                    interval_us=_RESPONSE_INTERVAL_US,
                )
            )

        # Invokes the actual RPC
        call = self._stub.StreamingOutputCall(request)

        self.assertFalse(call.cancelled())
        self.assertTrue(call.cancel())
        self.assertFalse(call.cancel())

        with self.assertRaises(asyncio.CancelledError):
            await call.read()

        self.assertTrue(call.cancelled())

        self.assertEqual(grpc.StatusCode.CANCELLED, await call.code())
        self.assertEqual(
            _LOCAL_CANCEL_DETAILS_EXPECTATION, await call.details()
        )

    async def test_late_cancel_unary_stream(self):
        """Test cancellation after received all messages."""
        # Prepares the request
        request = messages_pb2.StreamingOutputCallRequest()
        for _ in range(_NUM_STREAM_RESPONSES):
            request.response_parameters.append(
                messages_pb2.ResponseParameters(
                    size=_RESPONSE_PAYLOAD_SIZE,
                )
            )

        # Invokes the actual RPC
        call = self._stub.StreamingOutputCall(request)

        for _ in range(_NUM_STREAM_RESPONSES):
            response = await call.read()
            self.assertIs(
                type(response), messages_pb2.StreamingOutputCallResponse
            )
            self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))

        # After all messages received, it is possible that the final state
        # is received or on its way. It's basically a data race, so our
        # expectation here is do not crash :)
        call.cancel()
        self.assertIn(
            await call.code(), [grpc.StatusCode.OK, grpc.StatusCode.CANCELLED]
        )

    async def test_too_many_reads_unary_stream(self):
        """Test calling read after received all messages fails."""
        # Prepares the request
        request = messages_pb2.StreamingOutputCallRequest()
        for _ in range(_NUM_STREAM_RESPONSES):
            request.response_parameters.append(
                messages_pb2.ResponseParameters(
                    size=_RESPONSE_PAYLOAD_SIZE,
                )
            )

        # Invokes the actual RPC
        call = self._stub.StreamingOutputCall(request)

        for _ in range(_NUM_STREAM_RESPONSES):
            response = await call.read()
            self.assertIs(
                type(response), messages_pb2.StreamingOutputCallResponse
            )
            self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))
        self.assertIs(await call.read(), aio.EOF)

        # After the RPC is finished, further reads will lead to exception.
        self.assertEqual(await call.code(), grpc.StatusCode.OK)
        self.assertIs(await call.read(), aio.EOF)

    async def test_unary_stream_async_generator(self):
        """Sunny day test case for unary_stream."""
        # Prepares the request
        request = messages_pb2.StreamingOutputCallRequest()
        for _ in range(_NUM_STREAM_RESPONSES):
            request.response_parameters.append(
                messages_pb2.ResponseParameters(
                    size=_RESPONSE_PAYLOAD_SIZE,
                )
            )

        # Invokes the actual RPC
        call = self._stub.StreamingOutputCall(request)
        self.assertFalse(call.cancelled())

        async for response in call:
            self.assertIs(
                type(response), messages_pb2.StreamingOutputCallResponse
            )
            self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))

        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_cancel_unary_stream_with_many_interleavings(self):
        """A cheap alternative to a structured fuzzer.

        Certain classes of error only appear for very specific interleavings of
        coroutines. Rather than inserting semi-private asyncio.Events throughout
        the implementation on which to coordinate and explicilty waiting on those
        in tests, we instead search for bugs over the space of interleavings by
        stochastically varying the durations of certain events within the test.
        """

        # We range over several orders of magnitude to ensure that switching platforms
        # (i.e. to slow CI machines) does not result in this test becoming a no-op.
        sleep_ranges = (10.0**-i for i in range(1, 4))
        for sleep_range in sleep_ranges:
            for _ in range(_NONDETERMINISTIC_ITERATIONS):
                interval_us = random.randrange(
                    _NONDETERMINISTIC_SERVER_SLEEP_MAX_US
                )
                sleep_secs = sleep_range * random.random()

                coro_started = asyncio.Event()

                # Configs the server method to block forever
                request = messages_pb2.StreamingOutputCallRequest()
                request.response_parameters.append(
                    messages_pb2.ResponseParameters(
                        size=1,
                        interval_us=interval_us,
                    )
                )

                # Invokes the actual RPC
                call = self._stub.StreamingOutputCall(request)

                unhandled_error = False

                async def another_coro():
                    nonlocal unhandled_error
                    coro_started.set()
                    try:
                        await call.read()
                    except asyncio.CancelledError:
                        pass
                    except Exception as e:
                        unhandled_error = True
                        raise

                task = self.loop.create_task(another_coro())
                await coro_started.wait()
                await asyncio.sleep(sleep_secs)

                task.cancel()

                try:
                    await task
                except asyncio.CancelledError:
                    pass

                self.assertFalse(unhandled_error)

    async def test_cancel_unary_stream_in_task_using_read(self):
        coro_started = asyncio.Event()

        # Configs the server method to block forever
        request = messages_pb2.StreamingOutputCallRequest()
        request.response_parameters.append(
            messages_pb2.ResponseParameters(
                size=_RESPONSE_PAYLOAD_SIZE,
                interval_us=_INFINITE_INTERVAL_US,
            )
        )

        # Invokes the actual RPC
        call = self._stub.StreamingOutputCall(request)

        async def another_coro():
            coro_started.set()
            await call.read()

        task = self.loop.create_task(another_coro())
        await coro_started.wait()

        self.assertFalse(task.done())
        task.cancel()

        self.assertEqual(grpc.StatusCode.CANCELLED, await call.code())

        with self.assertRaises(asyncio.CancelledError):
            await task

    async def test_cancel_unary_stream_in_task_using_async_for(self):
        coro_started = asyncio.Event()

        # Configs the server method to block forever
        request = messages_pb2.StreamingOutputCallRequest()
        request.response_parameters.append(
            messages_pb2.ResponseParameters(
                size=_RESPONSE_PAYLOAD_SIZE,
                interval_us=_INFINITE_INTERVAL_US,
            )
        )

        # Invokes the actual RPC
        call = self._stub.StreamingOutputCall(request)

        async def another_coro():
            coro_started.set()
            async for _ in call:
                pass

        task = self.loop.create_task(another_coro())
        await coro_started.wait()

        self.assertFalse(task.done())
        task.cancel()

        self.assertEqual(grpc.StatusCode.CANCELLED, await call.code())

        with self.assertRaises(asyncio.CancelledError):
            await task

    async def test_time_remaining(self):
        request = messages_pb2.StreamingOutputCallRequest()
        # First message comes back immediately
        request.response_parameters.append(
            messages_pb2.ResponseParameters(
                size=_RESPONSE_PAYLOAD_SIZE,
            )
        )
        # Second message comes back after a unit of wait time
        request.response_parameters.append(
            messages_pb2.ResponseParameters(
                size=_RESPONSE_PAYLOAD_SIZE,
                interval_us=_RESPONSE_INTERVAL_US,
            )
        )

        call = self._stub.StreamingOutputCall(
            request, timeout=_SHORT_TIMEOUT_S * 2
        )

        response = await call.read()
        self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))

        # Should be around the same as the timeout
        remained_time = call.time_remaining()
        self.assertGreater(remained_time, _SHORT_TIMEOUT_S * 3 / 2)
        self.assertLess(remained_time, _SHORT_TIMEOUT_S * 5 / 2)

        response = await call.read()
        self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))

        # Should be around the timeout minus a unit of wait time
        remained_time = call.time_remaining()
        self.assertGreater(remained_time, _SHORT_TIMEOUT_S / 2)
        self.assertLess(remained_time, _SHORT_TIMEOUT_S * 3 / 2)

        self.assertEqual(grpc.StatusCode.OK, await call.code())

    async def test_empty_responses(self):
        # Prepares the request
        request = messages_pb2.StreamingOutputCallRequest()
        for _ in range(_NUM_STREAM_RESPONSES):
            request.response_parameters.append(
                messages_pb2.ResponseParameters()
            )

        # Invokes the actual RPC
        call = self._stub.StreamingOutputCall(request)

        for _ in range(_NUM_STREAM_RESPONSES):
            response = await call.read()
            self.assertIs(
                type(response), messages_pb2.StreamingOutputCallResponse
            )
            self.assertEqual(b"", response.SerializeToString())

        self.assertEqual(grpc.StatusCode.OK, await call.code())


class TestStreamUnaryCall(_MulticallableTestMixin, AioTestBase):
    async def test_cancel_stream_unary(self):
        call = self._stub.StreamingInputCall()

        # Prepares the request
        payload = messages_pb2.Payload(body=b"\0" * _REQUEST_PAYLOAD_SIZE)
        request = messages_pb2.StreamingInputCallRequest(payload=payload)

        # Sends out requests
        for _ in range(_NUM_STREAM_RESPONSES):
            await call.write(request)

        # Cancels the RPC
        self.assertFalse(call.done())
        self.assertFalse(call.cancelled())
        self.assertTrue(call.cancel())
        self.assertTrue(call.cancelled())

        await call.done_writing()

        with self.assertRaises(asyncio.CancelledError):
            await call

    async def test_early_cancel_stream_unary(self):
        call = self._stub.StreamingInputCall()

        # Cancels the RPC
        self.assertFalse(call.done())
        self.assertFalse(call.cancelled())
        self.assertTrue(call.cancel())
        self.assertTrue(call.cancelled())

        with self.assertRaises(asyncio.InvalidStateError):
            await call.write(messages_pb2.StreamingInputCallRequest())

        # Should be no-op
        await call.done_writing()

        with self.assertRaises(asyncio.CancelledError):
            await call

    async def test_write_after_done_writing(self):
        call = self._stub.StreamingInputCall()

        # Prepares the request
        payload = messages_pb2.Payload(body=b"\0" * _REQUEST_PAYLOAD_SIZE)
        request = messages_pb2.StreamingInputCallRequest(payload=payload)

        # Sends out requests
        for _ in range(_NUM_STREAM_RESPONSES):
            await call.write(request)

        # Should be no-op
        await call.done_writing()

        with self.assertRaises(asyncio.InvalidStateError):
            await call.write(messages_pb2.StreamingInputCallRequest())

        response = await call
        self.assertIsInstance(response, messages_pb2.StreamingInputCallResponse)
        self.assertEqual(
            _NUM_STREAM_RESPONSES * _REQUEST_PAYLOAD_SIZE,
            response.aggregated_payload_size,
        )

        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_error_in_async_generator(self):
        # Server will pause between responses
        request = messages_pb2.StreamingOutputCallRequest()
        request.response_parameters.append(
            messages_pb2.ResponseParameters(
                size=_RESPONSE_PAYLOAD_SIZE,
                interval_us=_RESPONSE_INTERVAL_US,
            )
        )

        # We expect the request iterator to receive the exception
        request_iterator_received_the_exception = asyncio.Event()

        async def request_iterator():
            with self.assertRaises(asyncio.CancelledError):
                for _ in range(_NUM_STREAM_RESPONSES):
                    yield request
                    await asyncio.sleep(_SHORT_TIMEOUT_S)
            request_iterator_received_the_exception.set()

        call = self._stub.StreamingInputCall(request_iterator())

        # Cancel the RPC after at least one response
        async def cancel_later():
            await asyncio.sleep(_SHORT_TIMEOUT_S * 2)
            call.cancel()

        cancel_later_task = self.loop.create_task(cancel_later())

        with self.assertRaises(asyncio.CancelledError):
            await call

        await request_iterator_received_the_exception.wait()

        # No failures in the cancel later task!
        await cancel_later_task

    async def test_normal_iterable_requests(self):
        # Prepares the request
        payload = messages_pb2.Payload(body=b"\0" * _REQUEST_PAYLOAD_SIZE)
        request = messages_pb2.StreamingInputCallRequest(payload=payload)
        requests = [request] * _NUM_STREAM_RESPONSES

        # Sends out requests
        call = self._stub.StreamingInputCall(requests)

        # RPC should succeed
        response = await call
        self.assertIsInstance(response, messages_pb2.StreamingInputCallResponse)
        self.assertEqual(
            _NUM_STREAM_RESPONSES * _REQUEST_PAYLOAD_SIZE,
            response.aggregated_payload_size,
        )

        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_call_rpc_error(self):
        async with aio.insecure_channel(UNREACHABLE_TARGET) as channel:
            stub = test_pb2_grpc.TestServiceStub(channel)

            # The error should be raised automatically without any traffic.
            call = stub.StreamingInputCall()
            with self.assertRaises(aio.AioRpcError) as exception_context:
                await call

            self.assertEqual(
                grpc.StatusCode.UNAVAILABLE, exception_context.exception.code()
            )

            self.assertTrue(call.done())
            self.assertEqual(grpc.StatusCode.UNAVAILABLE, await call.code())

    async def test_timeout(self):
        call = self._stub.StreamingInputCall(timeout=_SHORT_TIMEOUT_S)

        # The error should be raised automatically without any traffic.
        with self.assertRaises(aio.AioRpcError) as exception_context:
            await call

        rpc_error = exception_context.exception
        self.assertEqual(grpc.StatusCode.DEADLINE_EXCEEDED, rpc_error.code())
        self.assertTrue(call.done())
        self.assertEqual(grpc.StatusCode.DEADLINE_EXCEEDED, await call.code())


# Prepares the request that stream in a ping-pong manner.
_STREAM_OUTPUT_REQUEST_ONE_RESPONSE = messages_pb2.StreamingOutputCallRequest()
_STREAM_OUTPUT_REQUEST_ONE_RESPONSE.response_parameters.append(
    messages_pb2.ResponseParameters(size=_RESPONSE_PAYLOAD_SIZE)
)
_STREAM_OUTPUT_REQUEST_ONE_EMPTY_RESPONSE = (
    messages_pb2.StreamingOutputCallRequest()
)
_STREAM_OUTPUT_REQUEST_ONE_EMPTY_RESPONSE.response_parameters.append(
    messages_pb2.ResponseParameters()
)


class TestStreamStreamCall(_MulticallableTestMixin, AioTestBase):
    async def test_cancel(self):
        # Invokes the actual RPC
        call = self._stub.FullDuplexCall()

        for _ in range(_NUM_STREAM_RESPONSES):
            await call.write(_STREAM_OUTPUT_REQUEST_ONE_RESPONSE)
            response = await call.read()
            self.assertIsInstance(
                response, messages_pb2.StreamingOutputCallResponse
            )
            self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))

        # Cancels the RPC
        self.assertFalse(call.done())
        self.assertFalse(call.cancelled())
        self.assertTrue(call.cancel())
        self.assertTrue(call.cancelled())
        self.assertEqual(grpc.StatusCode.CANCELLED, await call.code())

    async def test_cancel_with_pending_read(self):
        call = self._stub.FullDuplexCall()

        await call.write(_STREAM_OUTPUT_REQUEST_ONE_RESPONSE)

        # Cancels the RPC
        self.assertFalse(call.done())
        self.assertFalse(call.cancelled())
        self.assertTrue(call.cancel())
        self.assertTrue(call.cancelled())
        self.assertEqual(grpc.StatusCode.CANCELLED, await call.code())

    async def test_cancel_with_ongoing_read(self):
        call = self._stub.FullDuplexCall()
        coro_started = asyncio.Event()

        async def read_coro():
            coro_started.set()
            await call.read()

        read_task = self.loop.create_task(read_coro())
        await coro_started.wait()
        self.assertFalse(read_task.done())

        # Cancels the RPC
        self.assertFalse(call.done())
        self.assertFalse(call.cancelled())
        self.assertTrue(call.cancel())
        self.assertTrue(call.cancelled())
        self.assertEqual(grpc.StatusCode.CANCELLED, await call.code())

    async def test_early_cancel(self):
        call = self._stub.FullDuplexCall()

        # Cancels the RPC
        self.assertFalse(call.done())
        self.assertFalse(call.cancelled())
        self.assertTrue(call.cancel())
        self.assertTrue(call.cancelled())
        self.assertEqual(grpc.StatusCode.CANCELLED, await call.code())

    async def test_cancel_after_done_writing(self):
        call = self._stub.FullDuplexCall()
        await call.done_writing()

        # Cancels the RPC
        self.assertFalse(call.done())
        self.assertFalse(call.cancelled())
        self.assertTrue(call.cancel())
        self.assertTrue(call.cancelled())
        self.assertEqual(grpc.StatusCode.CANCELLED, await call.code())

    async def test_late_cancel(self):
        call = self._stub.FullDuplexCall()
        await call.done_writing()
        self.assertEqual(grpc.StatusCode.OK, await call.code())

        # Cancels the RPC
        self.assertTrue(call.done())
        self.assertFalse(call.cancelled())
        self.assertFalse(call.cancel())
        self.assertFalse(call.cancelled())

        # Status is still OK
        self.assertEqual(grpc.StatusCode.OK, await call.code())

    async def test_async_generator(self):
        async def request_generator():
            yield _STREAM_OUTPUT_REQUEST_ONE_RESPONSE
            yield _STREAM_OUTPUT_REQUEST_ONE_RESPONSE

        call = self._stub.FullDuplexCall(request_generator())
        async for response in call:
            self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))

        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_too_many_reads(self):
        async def request_generator():
            for _ in range(_NUM_STREAM_RESPONSES):
                yield _STREAM_OUTPUT_REQUEST_ONE_RESPONSE

        call = self._stub.FullDuplexCall(request_generator())
        for _ in range(_NUM_STREAM_RESPONSES):
            response = await call.read()
            self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))
        self.assertIs(await call.read(), aio.EOF)

        self.assertEqual(await call.code(), grpc.StatusCode.OK)
        # After the RPC finished, the read should also produce EOF
        self.assertIs(await call.read(), aio.EOF)

    async def test_read_write_after_done_writing(self):
        call = self._stub.FullDuplexCall()

        # Writes two requests, and pending two requests
        await call.write(_STREAM_OUTPUT_REQUEST_ONE_RESPONSE)
        await call.write(_STREAM_OUTPUT_REQUEST_ONE_RESPONSE)
        await call.done_writing()

        # Further write should fail
        with self.assertRaises(asyncio.InvalidStateError):
            await call.write(_STREAM_OUTPUT_REQUEST_ONE_RESPONSE)

        # But read should be unaffected
        response = await call.read()
        self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))
        response = await call.read()
        self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))

        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_error_in_async_generator(self):
        # Server will pause between responses
        request = messages_pb2.StreamingOutputCallRequest()
        request.response_parameters.append(
            messages_pb2.ResponseParameters(
                size=_RESPONSE_PAYLOAD_SIZE,
                interval_us=_RESPONSE_INTERVAL_US,
            )
        )

        # We expect the request iterator to receive the exception
        request_iterator_received_the_exception = asyncio.Event()

        async def request_iterator():
            with self.assertRaises(asyncio.CancelledError):
                for _ in range(_NUM_STREAM_RESPONSES):
                    yield request
                    await asyncio.sleep(_SHORT_TIMEOUT_S)
            request_iterator_received_the_exception.set()

        call = self._stub.FullDuplexCall(request_iterator())

        # Cancel the RPC after at least one response
        async def cancel_later():
            await asyncio.sleep(_SHORT_TIMEOUT_S * 2)
            call.cancel()

        cancel_later_task = self.loop.create_task(cancel_later())

        with self.assertRaises(asyncio.CancelledError):
            async for response in call:
                self.assertEqual(
                    _RESPONSE_PAYLOAD_SIZE, len(response.payload.body)
                )

        await request_iterator_received_the_exception.wait()

        self.assertEqual(grpc.StatusCode.CANCELLED, await call.code())
        # No failures in the cancel later task!
        await cancel_later_task

    async def test_normal_iterable_requests(self):
        requests = [_STREAM_OUTPUT_REQUEST_ONE_RESPONSE] * _NUM_STREAM_RESPONSES

        call = self._stub.FullDuplexCall(iter(requests))
        async for response in call:
            self.assertEqual(_RESPONSE_PAYLOAD_SIZE, len(response.payload.body))

        self.assertEqual(await call.code(), grpc.StatusCode.OK)

    async def test_empty_ping_pong(self):
        call = self._stub.FullDuplexCall()
        for _ in range(_NUM_STREAM_RESPONSES):
            await call.write(_STREAM_OUTPUT_REQUEST_ONE_EMPTY_RESPONSE)
            response = await call.read()
            self.assertEqual(b"", response.SerializeToString())
        await call.done_writing()
        self.assertEqual(await call.code(), grpc.StatusCode.OK)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    unittest.main(verbosity=2)



#!/usr/bin/env python3
# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import multiprocessing
import os
import subprocess
import sys

sys.path.append(
    os.path.join(
        os.path.dirname(sys.argv[0]), "..", "run_tests", "python_utils"
    )
)
import jobset

clang_tidy = os.environ.get("CLANG_TIDY", "clang-tidy")

argp = argparse.ArgumentParser(description="Run clang-tidy against core")
argp.add_argument("files", nargs="+", help="Files to tidy")
argp.add_argument("--fix", dest="fix", action="store_true")
argp.add_argument(
    "-j",
    "--jobs",
    type=int,
    default=multiprocessing.cpu_count(),
    help="Number of CPUs to use",
)
argp.add_argument("--only-changed", dest="only_changed", action="store_true")
argp.set_defaults(fix=False, only_changed=False)
args = argp.parse_args()

# Explicitly passing the .clang-tidy config by reading it.
# This is required because source files in the compilation database are
# in a different source tree so clang-tidy cannot find the right config file
# by seeking their parent directories.
with open(".clang-tidy") as f:
    config = f.read()
cmdline = [
    clang_tidy,
    "--config=" + config,
]

if args.fix:
    cmdline.append("--fix-errors")

if args.only_changed:
    orig_files = set(args.files)
    actual_files = []
    output = subprocess.check_output(
        ["git", "diff", "upstream/master", "HEAD", "--name-only"]
    )
    for line in output.decode("ascii").splitlines(False):
        if line in orig_files:
            print(("check: %s" % line))
            actual_files.append(line)
        else:
            print(("skip: %s - not in the build" % line))
    args.files = actual_files

jobs = []
for filename in args.files:
    jobs.append(
        jobset.JobSpec(
            cmdline + [filename],
            shortname=filename,
            timeout_seconds=15 * 60,
        )
    )

num_fails, res_set = jobset.run(jobs, maxjobs=args.jobs, quiet_success=True)
sys.exit(num_fails)



# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform
import threading
import time
import unittest

from grpc._cython import cygrpc

from tests.unit import resources
from tests.unit import test_common
from tests.unit._cython import test_utilities

_SSL_HOST_OVERRIDE = b"foo.test.google.fr"
_CALL_CREDENTIALS_METADATA_KEY = "call-creds-key"
_CALL_CREDENTIALS_METADATA_VALUE = "call-creds-value"
_EMPTY_FLAGS = 0


def _metadata_plugin(context, callback):
    callback(
        (
            (
                _CALL_CREDENTIALS_METADATA_KEY,
                _CALL_CREDENTIALS_METADATA_VALUE,
            ),
        ),
        cygrpc.StatusCode.ok,
        b"",
    )


class TypeSmokeTest(unittest.TestCase):
    def testCompletionQueueUpDown(self):
        completion_queue = cygrpc.CompletionQueue()
        del completion_queue

    def testServerUpDown(self):
        server = cygrpc.Server(
            set(
                [
                    (
                        b"grpc.so_reuseport",
                        0,
                    )
                ]
            ),
            False,
        )
        del server

    def testChannelUpDown(self):
        channel = cygrpc.Channel(b"[::]:0", None, None)
        channel.close(cygrpc.StatusCode.cancelled, "Test method anyway!")

    def test_metadata_plugin_call_credentials_up_down(self):
        cygrpc.MetadataPluginCallCredentials(
            _metadata_plugin, b"test plugin name!"
        )

    def testServerStartNoExplicitShutdown(self):
        server = cygrpc.Server(
            [
                (
                    b"grpc.so_reuseport",
                    0,
                )
            ],
            False,
        )
        completion_queue = cygrpc.CompletionQueue()
        server.register_completion_queue(completion_queue)
        port = server.add_http2_port(b"[::]:0")
        self.assertIsInstance(port, int)
        server.start()
        del server

    def testServerStartShutdown(self):
        completion_queue = cygrpc.CompletionQueue()
        server = cygrpc.Server(
            [
                (
                    b"grpc.so_reuseport",
                    0,
                ),
            ],
            False,
        )
        server.add_http2_port(b"[::]:0")
        server.register_completion_queue(completion_queue)
        server.start()
        shutdown_tag = object()
        server.shutdown(completion_queue, shutdown_tag)
        event = completion_queue.poll()
        self.assertEqual(
            cygrpc.CompletionType.operation_complete, event.completion_type
        )
        self.assertIs(shutdown_tag, event.tag)
        del server
        del completion_queue


class ServerClientMixin(object):
    def setUpMixin(self, server_credentials, client_credentials, host_override):
        self.server_completion_queue = cygrpc.CompletionQueue()
        self.server = cygrpc.Server(
            [
                (
                    b"grpc.so_reuseport",
                    0,
                )
            ],
            False,
        )
        self.server.register_completion_queue(self.server_completion_queue)
        if server_credentials:
            self.port = self.server.add_http2_port(
                b"[::]:0", server_credentials
            )
        else:
            self.port = self.server.add_http2_port(b"[::]:0")
        self.server.start()
        self.client_completion_queue = cygrpc.CompletionQueue()
        if client_credentials:
            client_channel_arguments = (
                (
                    cygrpc.ChannelArgKey.ssl_target_name_override,
                    host_override,
                ),
            )
            self.client_channel = cygrpc.Channel(
                "localhost:{}".format(self.port).encode(),
                client_channel_arguments,
                client_credentials,
            )
        else:
            self.client_channel = cygrpc.Channel(
                "localhost:{}".format(self.port).encode(), set(), None
            )
        if host_override:
            self.host_argument = None  # default host
            self.expected_host = host_override
        else:
            # arbitrary host name necessitating no further identification
            self.host_argument = b"hostess"
            self.expected_host = self.host_argument

    def tearDownMixin(self):
        self.client_channel.close(cygrpc.StatusCode.ok, "test being torn down!")
        del self.client_channel
        del self.server
        del self.client_completion_queue
        del self.server_completion_queue

    def _perform_queue_operations(
        self, operations, call, queue, deadline, description
    ):
        """Perform the operations with given call, queue, and deadline.

        Invocation errors are reported with as an exception with `description`
        in the message. Performs the operations asynchronously, returning a
        future.
        """

        def performer():
            tag = object()
            try:
                call_result = call.start_client_batch(operations, tag)
                self.assertEqual(cygrpc.CallError.ok, call_result)
                event = queue.poll(deadline=deadline)
                self.assertEqual(
                    cygrpc.CompletionType.operation_complete,
                    event.completion_type,
                )
                self.assertTrue(event.success)
                self.assertIs(tag, event.tag)
            except Exception as error:
                raise Exception(
                    "Error in '{}': {}".format(description, error.message)
                )
            return event

        return test_utilities.SimpleFuture(performer)

    def test_echo(self):
        DEADLINE = time.time() + 5
        DEADLINE_TOLERANCE = 0.25
        CLIENT_METADATA_ASCII_KEY = "key"
        CLIENT_METADATA_ASCII_VALUE = "val"
        CLIENT_METADATA_BIN_KEY = "key-bin"
        CLIENT_METADATA_BIN_VALUE = b"\0" * 1000
        SERVER_INITIAL_METADATA_KEY = "init_me_me_me"
        SERVER_INITIAL_METADATA_VALUE = "whodawha?"
        SERVER_TRAILING_METADATA_KEY = "california_is_in_a_drought"
        SERVER_TRAILING_METADATA_VALUE = "zomg it is"
        SERVER_STATUS_CODE = cygrpc.StatusCode.ok
        SERVER_STATUS_DETAILS = "our work is never over"
        REQUEST = b"in death a member of project mayhem has a name"
        RESPONSE = b"his name is robert paulson"
        METHOD = b"twinkies"

        server_request_tag = object()
        request_call_result = self.server.request_call(
            self.server_completion_queue,
            self.server_completion_queue,
            server_request_tag,
        )

        self.assertEqual(cygrpc.CallError.ok, request_call_result)

        client_call_tag = object()
        client_initial_metadata = (
            (
                CLIENT_METADATA_ASCII_KEY,
                CLIENT_METADATA_ASCII_VALUE,
            ),
            (
                CLIENT_METADATA_BIN_KEY,
                CLIENT_METADATA_BIN_VALUE,
            ),
        )
        client_call = self.client_channel.integrated_call(
            0,
            METHOD,
            self.host_argument,
            DEADLINE,
            client_initial_metadata,
            None,
            [
                (
                    [
                        cygrpc.SendInitialMetadataOperation(
                            client_initial_metadata, _EMPTY_FLAGS
                        ),
                        cygrpc.SendMessageOperation(REQUEST, _EMPTY_FLAGS),
                        cygrpc.SendCloseFromClientOperation(_EMPTY_FLAGS),
                        cygrpc.ReceiveInitialMetadataOperation(_EMPTY_FLAGS),
                        cygrpc.ReceiveMessageOperation(_EMPTY_FLAGS),
                        cygrpc.ReceiveStatusOnClientOperation(_EMPTY_FLAGS),
                    ],
                    client_call_tag,
                ),
            ],
        )
        client_event_future = test_utilities.SimpleFuture(
            self.client_channel.next_call_event
        )

        request_event = self.server_completion_queue.poll(deadline=DEADLINE)
        self.assertEqual(
            cygrpc.CompletionType.operation_complete,
            request_event.completion_type,
        )
        self.assertIsInstance(request_event.call, cygrpc.Call)
        self.assertIs(server_request_tag, request_event.tag)
        self.assertTrue(
            test_common.metadata_transmitted(
                client_initial_metadata, request_event.invocation_metadata
            )
        )
        self.assertEqual(METHOD, request_event.call_details.method)
        self.assertEqual(self.expected_host, request_event.call_details.host)
        self.assertLess(
            abs(DEADLINE - request_event.call_details.deadline),
            DEADLINE_TOLERANCE,
        )

        server_call_tag = object()
        server_call = request_event.call
        server_start_batch_result = server_call.start_server_batch(
            [
                cygrpc.ReceiveMessageOperation(_EMPTY_FLAGS),
            ],
            server_call_tag,
        )
        self.assertEqual(cygrpc.CallError.ok, server_start_batch_result)

        server_message_event = self.server_completion_queue.poll(
            deadline=DEADLINE
        )

        server_call_tag = object()
        server_initial_metadata = (
            (
                SERVER_INITIAL_METADATA_KEY,
                SERVER_INITIAL_METADATA_VALUE,
            ),
        )
        server_trailing_metadata = (
            (
                SERVER_TRAILING_METADATA_KEY,
                SERVER_TRAILING_METADATA_VALUE,
            ),
        )
        server_start_batch_result = server_call.start_server_batch(
            [
                cygrpc.SendInitialMetadataOperation(
                    server_initial_metadata, _EMPTY_FLAGS
                ),
                cygrpc.SendMessageOperation(RESPONSE, _EMPTY_FLAGS),
                cygrpc.ReceiveCloseOnServerOperation(_EMPTY_FLAGS),
                cygrpc.SendStatusFromServerOperation(
                    server_trailing_metadata,
                    SERVER_STATUS_CODE,
                    SERVER_STATUS_DETAILS,
                    _EMPTY_FLAGS,
                ),
            ],
            server_call_tag,
        )
        self.assertEqual(cygrpc.CallError.ok, server_start_batch_result)

        server_event = self.server_completion_queue.poll(deadline=DEADLINE)
        client_event = client_event_future.result()

        self.assertEqual(6, len(client_event.batch_operations))
        found_client_op_types = set()
        for client_result in client_event.batch_operations:
            # we expect each op type to be unique
            self.assertNotIn(client_result.type(), found_client_op_types)
            found_client_op_types.add(client_result.type())
            if (
                client_result.type()
                == cygrpc.OperationType.receive_initial_metadata
            ):
                self.assertTrue(
                    test_common.metadata_transmitted(
                        server_initial_metadata,
                        client_result.initial_metadata(),
                    )
                )
            elif client_result.type() == cygrpc.OperationType.receive_message:
                self.assertEqual(RESPONSE, client_result.message())
            elif (
                client_result.type()
                == cygrpc.OperationType.receive_status_on_client
            ):
                self.assertTrue(
                    test_common.metadata_transmitted(
                        server_trailing_metadata,
                        client_result.trailing_metadata(),
                    )
                )
                self.assertEqual(SERVER_STATUS_DETAILS, client_result.details())
                self.assertEqual(SERVER_STATUS_CODE, client_result.code())
        self.assertEqual(
            set(
                [
                    cygrpc.OperationType.send_initial_metadata,
                    cygrpc.OperationType.send_message,
                    cygrpc.OperationType.send_close_from_client,
                    cygrpc.OperationType.receive_initial_metadata,
                    cygrpc.OperationType.receive_message,
                    cygrpc.OperationType.receive_status_on_client,
                ]
            ),
            found_client_op_types,
        )

        self.assertEqual(1, len(server_message_event.batch_operations))
        found_server_op_types = set()
        for server_result in server_message_event.batch_operations:
            self.assertNotIn(server_result.type(), found_server_op_types)
            found_server_op_types.add(server_result.type())
            if server_result.type() == cygrpc.OperationType.receive_message:
                self.assertEqual(REQUEST, server_result.message())
            elif (
                server_result.type()
                == cygrpc.OperationType.receive_close_on_server
            ):
                self.assertFalse(server_result.cancelled())
        self.assertEqual(
            set(
                [
                    cygrpc.OperationType.receive_message,
                ]
            ),
            found_server_op_types,
        )

        self.assertEqual(4, len(server_event.batch_operations))
        found_server_op_types = set()
        for server_result in server_event.batch_operations:
            self.assertNotIn(server_result.type(), found_server_op_types)
            found_server_op_types.add(server_result.type())
            if server_result.type() == cygrpc.OperationType.receive_message:
                self.assertEqual(REQUEST, server_result.message())
            elif (
                server_result.type()
                == cygrpc.OperationType.receive_close_on_server
            ):
                self.assertFalse(server_result.cancelled())
        self.assertEqual(
            set(
                [
                    cygrpc.OperationType.send_initial_metadata,
                    cygrpc.OperationType.send_message,
                    cygrpc.OperationType.receive_close_on_server,
                    cygrpc.OperationType.send_status_from_server,
                ]
            ),
            found_server_op_types,
        )

        del client_call
        del server_call

    def test_6522(self):
        DEADLINE = time.time() + 5
        DEADLINE_TOLERANCE = 0.25
        METHOD = b"twinkies"

        empty_metadata = ()

        # Prologue
        server_request_tag = object()
        self.server.request_call(
            self.server_completion_queue,
            self.server_completion_queue,
            server_request_tag,
        )
        client_call = self.client_channel.segregated_call(
            0,
            METHOD,
            self.host_argument,
            DEADLINE,
            None,
            None,
            (
                [
                    (
                        [
                            cygrpc.SendInitialMetadataOperation(
                                empty_metadata, _EMPTY_FLAGS
                            ),
                            cygrpc.ReceiveInitialMetadataOperation(
                                _EMPTY_FLAGS
                            ),
                        ],
                        object(),
                    ),
                    (
                        [
                            cygrpc.ReceiveStatusOnClientOperation(_EMPTY_FLAGS),
                        ],
                        object(),
                    ),
                ]
            ),
        )

        client_initial_metadata_event_future = test_utilities.SimpleFuture(
            client_call.next_event
        )

        request_event = self.server_completion_queue.poll(deadline=DEADLINE)
        server_call = request_event.call

        def perform_server_operations(operations, description):
            return self._perform_queue_operations(
                operations,
                server_call,
                self.server_completion_queue,
                DEADLINE,
                description,
            )

        server_event_future = perform_server_operations(
            [
                cygrpc.SendInitialMetadataOperation(
                    empty_metadata, _EMPTY_FLAGS
                ),
            ],
            "Server prologue",
        )

        client_initial_metadata_event_future.result()  # force completion
        server_event_future.result()

        # Messaging
        for _ in range(10):
            client_call.operate(
                [
                    cygrpc.SendMessageOperation(b"", _EMPTY_FLAGS),
                    cygrpc.ReceiveMessageOperation(_EMPTY_FLAGS),
                ],
                "Client message",
            )
            client_message_event_future = test_utilities.SimpleFuture(
                client_call.next_event
            )
            server_event_future = perform_server_operations(
                [
                    cygrpc.SendMessageOperation(b"", _EMPTY_FLAGS),
                    cygrpc.ReceiveMessageOperation(_EMPTY_FLAGS),
                ],
                "Server receive",
            )

            client_message_event_future.result()  # force completion
            server_event_future.result()

        # Epilogue
        client_call.operate(
            [
                cygrpc.SendCloseFromClientOperation(_EMPTY_FLAGS),
            ],
            "Client epilogue",
        )
        # One for ReceiveStatusOnClient, one for SendCloseFromClient.
        client_events_future = test_utilities.SimpleFuture(
            lambda: {
                client_call.next_event(),
                client_call.next_event(),
            }
        )

        server_event_future = perform_server_operations(
            [
                cygrpc.ReceiveCloseOnServerOperation(_EMPTY_FLAGS),
                cygrpc.SendStatusFromServerOperation(
                    empty_metadata, cygrpc.StatusCode.ok, b"", _EMPTY_FLAGS
                ),
            ],
            "Server epilogue",
        )

        client_events_future.result()  # force completion
        server_event_future.result()


class InsecureServerInsecureClient(unittest.TestCase, ServerClientMixin):
    def setUp(self):
        self.setUpMixin(None, None, None)

    def tearDown(self):
        self.tearDownMixin()


class SecureServerSecureClient(unittest.TestCase, ServerClientMixin):
    def setUp(self):
        server_credentials = cygrpc.server_credentials_ssl(
            None,
            [
                cygrpc.SslPemKeyCertPair(
                    resources.private_key(), resources.certificate_chain()
                )
            ],
            False,
        )
        client_credentials = cygrpc.SSLChannelCredentials(
            resources.test_root_certificates(), None, None
        )
        self.setUpMixin(
            server_credentials, client_credentials, _SSL_HOST_OVERRIDE
        )

    def tearDown(self):
        self.tearDownMixin()


if __name__ == "__main__":
    unittest.main(verbosity=2)



""" Multicast DNS Service Discovery for Python, v0.14-wmcbrine
    Copyright 2003 Paul Scott-Murphy, 2014 William McBrine

    This module provides a framework for the use of DNS Service Discovery
    using IP multicast.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
    USA
"""

import asyncio
import logging
import random
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union, cast

from ._logger import QuietLogger, log
from ._protocol.incoming import DNSIncoming
from ._transport import _WrappedTransport, make_wrapped_transport
from ._utils.time import current_time_millis, millis_to_seconds
from .const import _DUPLICATE_PACKET_SUPPRESSION_INTERVAL, _MAX_MSG_ABSOLUTE

if TYPE_CHECKING:
    from ._core import Zeroconf

_TC_DELAY_RANDOM_INTERVAL = (400, 500)


_bytes = bytes
_str = str
_int = int

logging_DEBUG = logging.DEBUG


class AsyncListener:

    """A Listener is used by this module to listen on the multicast
    group to which DNS messages are sent, allowing the implementation
    to cache information as it arrives.

    It requires registration with an Engine object in order to have
    the read() method called when a socket is available for reading."""

    __slots__ = (
        'zc',
        '_record_manager',
        'data',
        'last_time',
        'last_message',
        'transport',
        'sock_description',
        '_deferred',
        '_timers',
    )

    def __init__(self, zc: 'Zeroconf') -> None:
        self.zc = zc
        self._record_manager = zc.record_manager
        self.data: Optional[bytes] = None
        self.last_time: float = 0
        self.last_message: Optional[DNSIncoming] = None
        self.transport: Optional[_WrappedTransport] = None
        self.sock_description: Optional[str] = None
        self._deferred: Dict[str, List[DNSIncoming]] = {}
        self._timers: Dict[str, asyncio.TimerHandle] = {}
        super().__init__()

    def datagram_received(
        self, data: _bytes, addrs: Union[Tuple[str, int], Tuple[str, int, int, int]]
    ) -> None:
        assert self.transport is not None
        data_len = len(data)
        debug = log.isEnabledFor(logging_DEBUG)

        if data_len > _MAX_MSG_ABSOLUTE:
            # Guard against oversized packets to ensure bad implementations cannot overwhelm
            # the system.
            if debug:
                log.debug(
                    "Discarding incoming packet with length %s, which is larger "
                    "than the absolute maximum size of %s",
                    data_len,
                    _MAX_MSG_ABSOLUTE,
                )
            return

        now = current_time_millis()
        if (
            self.data == data
            and (now - _DUPLICATE_PACKET_SUPPRESSION_INTERVAL) < self.last_time
            and self.last_message is not None
            and not self.last_message.has_qu_question()
        ):
            # Guard against duplicate packets
            if debug:
                log.debug(
                    'Ignoring duplicate message with no unicast questions received from %s [socket %s] (%d bytes) as [%r]',
                    addrs,
                    self.sock_description,
                    data_len,
                    data,
                )
            return

        if len(addrs) == 2:
            v6_flow_scope: Union[Tuple[()], Tuple[int, int]] = ()
            # https://github.com/python/mypy/issues/1178
            addr, port = addrs  # type: ignore
            addr_port = addrs
            if TYPE_CHECKING:
                addr_port = cast(Tuple[str, int], addr_port)
            scope = None
        else:
            # https://github.com/python/mypy/issues/1178
            addr, port, flow, scope = addrs  # type: ignore
            if debug:  # pragma: no branch
                log.debug('IPv6 scope_id %d associated to the receiving interface', scope)
            v6_flow_scope = (flow, scope)
            addr_port = (addr, port)

        msg = DNSIncoming(data, addr_port, scope, now)
        self.data = data
        self.last_time = now
        self.last_message = msg
        if msg.valid:
            if debug:
                log.debug(
                    'Received from %r:%r [socket %s]: %r (%d bytes) as [%r]',
                    addr,
                    port,
                    self.sock_description,
                    msg,
                    data_len,
                    data,
                )
        else:
            if debug:
                log.debug(
                    'Received from %r:%r [socket %s]: (%d bytes) [%r]',
                    addr,
                    port,
                    self.sock_description,
                    data_len,
                    data,
                )
            return

        if not msg.is_query():
            self._record_manager.async_updates_from_response(msg)
            return

        self.handle_query_or_defer(msg, addr, port, self.transport, v6_flow_scope)

    def handle_query_or_defer(
        self,
        msg: DNSIncoming,
        addr: str,
        port: int,
        transport: _WrappedTransport,
        v6_flow_scope: Union[Tuple[()], Tuple[int, int]] = (),
    ) -> None:
        """Deal with incoming query packets.  Provides a response if
        possible."""
        if not msg.truncated:
            self._respond_query(msg, addr, port, transport, v6_flow_scope)
            return

        deferred = self._deferred.setdefault(addr, [])
        # If we get the same packet we ignore it
        for incoming in reversed(deferred):
            if incoming.data == msg.data:
                return
        deferred.append(msg)
        delay = millis_to_seconds(random.randint(*_TC_DELAY_RANDOM_INTERVAL))
        loop = self.zc.loop
        assert loop is not None
        self._cancel_any_timers_for_addr(addr)
        self._timers[addr] = loop.call_at(
            loop.time() + delay, self._respond_query, None, addr, port, transport, v6_flow_scope
        )

    def _cancel_any_timers_for_addr(self, addr: _str) -> None:
        """Cancel any future truncated packet timers for the address."""
        if addr in self._timers:
            self._timers.pop(addr).cancel()

    def _respond_query(
        self,
        msg: Optional[DNSIncoming],
        addr: _str,
        port: _int,
        transport: _WrappedTransport,
        v6_flow_scope: Union[Tuple[()], Tuple[int, int]] = (),
    ) -> None:
        """Respond to a query and reassemble any truncated deferred packets."""
        self._cancel_any_timers_for_addr(addr)
        packets = self._deferred.pop(addr, [])
        if msg:
            packets.append(msg)

        self.zc.handle_assembled_query(packets, addr, port, transport, v6_flow_scope)

    def error_received(self, exc: Exception) -> None:
        """Likely socket closed or IPv6."""
        # We preformat the message string with the socket as we want
        # log_exception_once to log a warrning message once PER EACH
        # different socket in case there are problems with multiple
        # sockets
        msg_str = f"Error with socket {self.sock_description}): %s"
        QuietLogger.log_exception_once(exc, msg_str, exc)

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        wrapped_transport = make_wrapped_transport(cast(asyncio.DatagramTransport, transport))
        self.transport = wrapped_transport
        self.sock_description = f"{wrapped_transport.fileno} ({wrapped_transport.sock_name})"

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Handle connection lost."""



# Copyright 2020 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import contextlib
import functools
import json
import logging
from typing import Any, Dict, List, Optional

from absl import flags
from google.cloud import secretmanager_v1
from google.longrunning import operations_pb2
from google.protobuf import json_format
from google.rpc import code_pb2
from google.rpc import error_details_pb2
from google.rpc import status_pb2
from googleapiclient import discovery
import googleapiclient.errors
import googleapiclient.http
import tenacity
import yaml

import framework.helpers.highlighter

logger = logging.getLogger(__name__)
PRIVATE_API_KEY_SECRET_NAME = flags.DEFINE_string(
    "private_api_key_secret_name",
    default=None,
    help=(
        "Load Private API access key from the latest version of the secret "
        "with the given name, in the format projects/*/secrets/*"
    ),
)
V1_DISCOVERY_URI = flags.DEFINE_string(
    "v1_discovery_uri",
    default=discovery.V1_DISCOVERY_URI,
    help="Override v1 Discovery URI",
)
V2_DISCOVERY_URI = flags.DEFINE_string(
    "v2_discovery_uri",
    default=discovery.V2_DISCOVERY_URI,
    help="Override v2 Discovery URI",
)
COMPUTE_V1_DISCOVERY_FILE = flags.DEFINE_string(
    "compute_v1_discovery_file",
    default=None,
    help="Load compute v1 from discovery file",
)
GCP_UI_URL = flags.DEFINE_string(
    "gcp_ui_url",
    default="console.cloud.google.com",
    help="Override GCP UI URL.",
)

# Type aliases
_HttpError = googleapiclient.errors.HttpError
_HttpLib2Error = googleapiclient.http.httplib2.HttpLib2Error
_HighlighterYaml = framework.helpers.highlighter.HighlighterYaml
Operation = operations_pb2.Operation
HttpRequest = googleapiclient.http.HttpRequest


class GcpApiManager:
    def __init__(
        self,
        *,
        v1_discovery_uri=None,
        v2_discovery_uri=None,
        compute_v1_discovery_file=None,
        private_api_key_secret_name=None,
        gcp_ui_url=None,
    ):
        self.v1_discovery_uri = v1_discovery_uri or V1_DISCOVERY_URI.value
        self.v2_discovery_uri = v2_discovery_uri or V2_DISCOVERY_URI.value
        self.compute_v1_discovery_file = (
            compute_v1_discovery_file or COMPUTE_V1_DISCOVERY_FILE.value
        )
        self.private_api_key_secret_name = (
            private_api_key_secret_name or PRIVATE_API_KEY_SECRET_NAME.value
        )
        self.gcp_ui_url = gcp_ui_url or GCP_UI_URL.value
        # TODO(sergiitk): add options to pass google Credentials
        self._exit_stack = contextlib.ExitStack()

    def close(self):
        self._exit_stack.close()

    @property
    @functools.lru_cache(None)
    def private_api_key(self):
        """
        Private API key.

        Return API key credential that identifies a GCP project allow-listed for
        accessing private API discovery documents.
        https://console.cloud.google.com/apis/credentials

        This method lazy-loads the content of the key from the Secret Manager.
        https://console.cloud.google.com/security/secret-manager
        """
        if not self.private_api_key_secret_name:
            raise ValueError(
                "private_api_key_secret_name must be set to "
                "access private_api_key."
            )

        secrets_api = self.secrets("v1")
        version_resource_path = secrets_api.secret_version_path(
            **secrets_api.parse_secret_path(self.private_api_key_secret_name),
            secret_version="latest",
        )
        secret: secretmanager_v1.AccessSecretVersionResponse
        secret = secrets_api.access_secret_version(name=version_resource_path)
        return secret.payload.data.decode()

    @functools.lru_cache(None)
    def compute(self, version):
        api_name = "compute"
        if version == "v1":
            if self.compute_v1_discovery_file:
                return self._build_from_file(self.compute_v1_discovery_file)
            else:
                return self._build_from_discovery_v1(api_name, version)
        elif version == "v1alpha":
            return self._build_from_discovery_v1(api_name, "alpha")

        raise NotImplementedError(f"Compute {version} not supported")

    @functools.lru_cache(None)
    def networksecurity(self, version):
        api_name = "networksecurity"
        if version == "v1alpha1":
            return self._build_from_discovery_v2(
                api_name,
                version,
                api_key=self.private_api_key,
                visibility_labels=["NETWORKSECURITY_ALPHA"],
            )
        elif version == "v1beta1":
            return self._build_from_discovery_v2(api_name, version)

        raise NotImplementedError(f"Network Security {version} not supported")

    @functools.lru_cache(None)
    def networkservices(self, version):
        api_name = "networkservices"
        if version == "v1alpha1":
            return self._build_from_discovery_v2(
                api_name,
                version,
                api_key=self.private_api_key,
                visibility_labels=["NETWORKSERVICES_ALPHA"],
            )
        elif version == "v1beta1":
            return self._build_from_discovery_v2(api_name, version)

        raise NotImplementedError(f"Network Services {version} not supported")

    @staticmethod
    @functools.lru_cache(None)
    def secrets(version: str):
        if version == "v1":
            return secretmanager_v1.SecretManagerServiceClient()

        raise NotImplementedError(f"Secret Manager {version} not supported")

    @functools.lru_cache(None)
    def iam(self, version: str) -> discovery.Resource:
        """Identity and Access Management (IAM) API.

        https://cloud.google.com/iam/docs/reference/rest
        https://googleapis.github.io/google-api-python-client/docs/dyn/iam_v1.html
        """
        api_name = "iam"
        if version == "v1":
            return self._build_from_discovery_v1(api_name, version)

        raise NotImplementedError(
            f"Identity and Access Management (IAM) {version} not supported"
        )

    def _build_from_discovery_v1(self, api_name, version):
        api = discovery.build(
            api_name,
            version,
            cache_discovery=False,
            discoveryServiceUrl=self.v1_discovery_uri,
        )
        self._exit_stack.enter_context(api)
        return api

    def _build_from_discovery_v2(
        self,
        api_name,
        version,
        *,
        api_key: Optional[str] = None,
        visibility_labels: Optional[List] = None,
    ):
        params = {}
        if api_key:
            params["key"] = api_key
        if visibility_labels:
            # Dash-separated list of labels.
            params["labels"] = "_".join(visibility_labels)

        params_str = ""
        if params:
            params_str = "&" + "&".join(f"{k}={v}" for k, v in params.items())

        api = discovery.build(
            api_name,
            version,
            cache_discovery=False,
            discoveryServiceUrl=f"{self.v2_discovery_uri}{params_str}",
        )
        self._exit_stack.enter_context(api)
        return api

    def _build_from_file(self, discovery_file):
        with open(discovery_file, "r") as f:
            api = discovery.build_from_document(f.read())
        self._exit_stack.enter_context(api)
        return api


class Error(Exception):
    """Base error class for GCP API errors."""


class ResponseError(Error):
    """The response was not a 2xx."""

    reason: str
    uri: str
    error_details: Optional[str]
    status: Optional[int]
    cause: _HttpError

    def __init__(self, cause: _HttpError):
        # TODO(sergiitk): cleanup when we upgrade googleapiclient:
        #  - remove _get_reason()
        #  - remove error_details note
        #  - use status_code()
        self.reason = cause._get_reason().strip()  # noqa
        self.uri = cause.uri
        self.error_details = cause.error_details  # NOTE: Must after _get_reason
        self.status = None
        if cause.resp and cause.resp.status:
            self.status = cause.resp.status
        self.cause = cause
        super().__init__()

    def __repr__(self):
        return (
            f"<ResponseError {self.status} when requesting {self.uri} "
            f'returned "{self.reason}". Details: "{self.error_details}">'
        )


class TransportError(Error):
    """A transport error has occurred."""

    cause: _HttpLib2Error

    def __init__(self, cause: _HttpLib2Error):
        self.cause = cause
        super().__init__()

    def __repr__(self):
        return f"<TransportError cause: {self.cause!r}>"


class OperationError(Error):
    """
    Operation was not successful.

    Assuming Operation based on Google API Style Guide:
    https://cloud.google.com/apis/design/design_patterns#long_running_operations
    https://github.com/googleapis/googleapis/blob/master/google/longrunning/operations.proto
    """

    api_name: str
    name: str
    metadata: Any
    code_name: code_pb2.Code
    error: status_pb2.Status

    def __init__(self, api_name: str, response: dict):
        self.api_name = api_name

        # Operation.metadata field is Any specific to the API. It may not be
        # present in the default descriptor pool, and that's expected.
        # To avoid json_format.ParseError, handle it separately.
        self.metadata = response.pop("metadata", {})

        # Must be after removing metadata field.
        operation: Operation = self._parse_operation_response(response)
        self.name = operation.name or "unknown"
        self.code_name = code_pb2.Code.Name(operation.error.code)
        self.error = operation.error
        super().__init__()

    @staticmethod
    def _parse_operation_response(operation_response: dict) -> Operation:
        try:
            return json_format.ParseDict(
                operation_response,
                Operation(),
                ignore_unknown_fields=True,
                descriptor_pool=error_details_pb2.DESCRIPTOR.pool,
            )
        except (json_format.Error, TypeError) as e:
            # Swallow parsing errors if any. Building correct OperationError()
            # is more important than losing debug information. Details still
            # can be extracted from the warning.
            logger.warning(
                (
                    "Can't parse response while processing OperationError:"
                    " '%r', error %r"
                ),
                operation_response,
                e,
            )
            return Operation()

    def __str__(self):
        indent_l1 = " " * 2
        indent_l2 = indent_l1 * 2

        result = (
            f'{self.api_name} operation "{self.name}" failed.\n'
            f"{indent_l1}code: {self.error.code} ({self.code_name})\n"
            f'{indent_l1}message: "{self.error.message}"'
        )

        if self.error.details:
            result += f"\n{indent_l1}details: [\n"
            for any_error in self.error.details:
                error_str = json_format.MessageToJson(any_error)
                for line in error_str.splitlines():
                    result += indent_l2 + line + "\n"
            result += f"{indent_l1}]"

        if self.metadata:
            result += f"\n  metadata: \n"
            metadata_str = json.dumps(self.metadata, indent=2)
            for line in metadata_str.splitlines():
                result += indent_l2 + line + "\n"
            result = result.rstrip()

        return result


class GcpProjectApiResource:
    # TODO(sergiitk): move someplace better
    _WAIT_FOR_OPERATION_SEC = 60 * 10
    _WAIT_FIXED_SEC = 2
    _GCP_API_RETRIES = 5

    def __init__(self, api: discovery.Resource, project: str):
        self.api: discovery.Resource = api
        self.project: str = project
        self._highlighter = _HighlighterYaml()

    # TODO(sergiitk): in upcoming GCP refactoring, differentiate between
    #   _execute for LRO (Long Running Operations), and immediate operations.
    def _execute(
        self,
        request: HttpRequest,
        *,
        num_retries: Optional[int] = _GCP_API_RETRIES,
    ) -> Dict[str, Any]:
        """Execute the immediate request.

        Returns:
          Unmarshalled response as a dictionary.

        Raises:
          ResponseError if the response was not a 2xx.
          TransportError if a transport error has occurred.
        """
        if num_retries is None:
            num_retries = self._GCP_API_RETRIES
        try:
            return request.execute(num_retries=num_retries)
        except _HttpError as error:
            raise ResponseError(error)
        except _HttpLib2Error as error:
            raise TransportError(error)

    def resource_pretty_format(
        self,
        resource: Any,
        *,
        highlight: bool = True,
    ) -> str:
        """Return a string with pretty-printed resource body."""
        yaml_out: str = yaml.dump(
            resource,
            explicit_start=True,
            explicit_end=True,
        )
        return self._highlighter.highlight(yaml_out) if highlight else yaml_out

    def resources_pretty_format(
        self,
        resources: list[Any],
        *,
        highlight: bool = True,
    ) -> str:
        out = []
        for resource in resources:
            if hasattr(resource, "name"):
                out.append(f"{resource.name}:")
            elif "name" in resource:
                out.append(f"{resource['name']}:")
            out.append(
                self.resource_pretty_format(resource, highlight=highlight)
            )
        return "\n".join(out)

    @staticmethod
    def wait_for_operation(
        operation_request,
        test_success_fn,
        timeout_sec=_WAIT_FOR_OPERATION_SEC,
        wait_sec=_WAIT_FIXED_SEC,
    ):
        retryer = tenacity.Retrying(
            retry=(
                tenacity.retry_if_not_result(test_success_fn)
                | tenacity.retry_if_exception_type()
            ),
            wait=tenacity.wait_fixed(wait_sec),
            stop=tenacity.stop_after_delay(timeout_sec),
            after=tenacity.after_log(logger, logging.DEBUG),
            reraise=True,
        )
        return retryer(operation_request.execute)


class GcpStandardCloudApiResource(GcpProjectApiResource, metaclass=abc.ABCMeta):
    GLOBAL_LOCATION = "global"

    def parent(self, location: Optional[str] = GLOBAL_LOCATION):
        if location is None:
            location = self.GLOBAL_LOCATION
        return f"projects/{self.project}/locations/{location}"

    def resource_full_name(self, name, collection_name):
        return f"{self.parent()}/{collection_name}/{name}"

    def _create_resource(
        self, collection: discovery.Resource, body: dict, **kwargs
    ):
        logger.info(
            "Creating %s resource:\n%s",
            self.api_name,
            self.resource_pretty_format(body),
        )
        create_req = collection.create(
            parent=self.parent(), body=body, **kwargs
        )
        self._execute(create_req)

    @property
    @abc.abstractmethod
    def api_name(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def api_version(self) -> str:
        raise NotImplementedError

    def _get_resource(self, collection: discovery.Resource, full_name):
        resource = collection.get(name=full_name).execute()
        logger.info(
            "Loaded %s:\n%s", full_name, self.resource_pretty_format(resource)
        )
        return resource

    def _delete_resource(
        self, collection: discovery.Resource, full_name: str
    ) -> bool:
        logger.debug("Deleting %s", full_name)
        try:
            self._execute(collection.delete(name=full_name))
            return True
        except _HttpError as error:
            if error.resp and error.resp.status == 404:
                logger.info("%s not deleted since it does not exist", full_name)
            else:
                logger.warning("Failed to delete %s, %r", full_name, error)
        return False

    # TODO(sergiitk): Use ResponseError and TransportError
    def _execute(  # pylint: disable=arguments-differ
        self,
        request: HttpRequest,
        timeout_sec: int = GcpProjectApiResource._WAIT_FOR_OPERATION_SEC,
    ):
        operation = request.execute(num_retries=self._GCP_API_RETRIES)
        logger.debug("Operation %s", operation)
        self._wait(operation["name"], timeout_sec)

    def _wait(
        self,
        operation_id: str,
        timeout_sec: int = GcpProjectApiResource._WAIT_FOR_OPERATION_SEC,
    ):
        logger.info(
            "Waiting %s sec for %s operation id: %s",
            timeout_sec,
            self.api_name,
            operation_id,
        )

        op_request = (
            self.api.projects().locations().operations().get(name=operation_id)
        )
        operation = self.wait_for_operation(
            operation_request=op_request,
            test_success_fn=lambda result: result["done"],
            timeout_sec=timeout_sec,
        )

        logger.debug("Completed operation: %s", operation)
        if "error" in operation:
            raise OperationError(self.api_name, operation)



# Copyright 2018 The gRPC Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent import futures
import logging
import unittest

import grpc

from tests.unit import resources


class _ActualGenericRpcHandler(grpc.GenericRpcHandler):
    def service(self, handler_call_details):
        return None


class ServerTest(unittest.TestCase):
    def test_not_a_generic_rpc_handler_at_construction(self):
        with self.assertRaises(AttributeError) as exception_context:
            grpc.server(
                futures.ThreadPoolExecutor(max_workers=5),
                handlers=[
                    _ActualGenericRpcHandler(),
                    object(),
                ],
            )
        self.assertIn(
            "grpc.GenericRpcHandler", str(exception_context.exception)
        )

    def test_not_a_generic_rpc_handler_after_construction(self):
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=5))
        with self.assertRaises(AttributeError) as exception_context:
            server.add_generic_rpc_handlers(
                [
                    _ActualGenericRpcHandler(),
                    object(),
                ]
            )
        self.assertIn(
            "grpc.GenericRpcHandler", str(exception_context.exception)
        )

    def test_failed_port_binding_exception(self):
        server = grpc.server(None, options=(("grpc.so_reuseport", 0),))
        port = server.add_insecure_port("localhost:0")
        bind_address = "localhost:%d" % port

        with self.assertRaises(RuntimeError):
            server.add_insecure_port(bind_address)

        server_credentials = grpc.ssl_server_credentials(
            [(resources.private_key(), resources.certificate_chain())]
        )
        with self.assertRaises(RuntimeError):
            server.add_secure_port(bind_address, server_credentials)


if __name__ == "__main__":
    logging.basicConfig()
    unittest.main(verbosity=2)



# Copyright 2021 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import List

from absl import flags
from absl.testing import absltest

from framework import xds_k8s_testcase

logger = logging.getLogger(__name__)
flags.adopt_module_key_flags(xds_k8s_testcase)

# Type aliases
_XdsTestServer = xds_k8s_testcase.XdsTestServer
_XdsTestClient = xds_k8s_testcase.XdsTestClient


class RoundRobinTest(xds_k8s_testcase.RegularXdsKubernetesTestCase):
    def test_round_robin(self) -> None:
        REPLICA_COUNT = 2

        with self.subTest("00_create_health_check"):
            self.td.create_health_check()

        with self.subTest("01_create_backend_services"):
            self.td.create_backend_service()

        with self.subTest("02_create_url_map"):
            self.td.create_url_map(self.server_xds_host, self.server_xds_port)

        with self.subTest("03_create_target_proxy"):
            self.td.create_target_proxy()

        with self.subTest("04_create_forwarding_rule"):
            self.td.create_forwarding_rule(self.server_xds_port)

        test_servers: List[_XdsTestServer]
        with self.subTest("05_start_test_servers"):
            test_servers = self.startTestServers(replica_count=REPLICA_COUNT)

        with self.subTest("06_add_server_backends_to_backend_services"):
            self.setupServerBackends()

        test_client: _XdsTestClient
        with self.subTest("07_start_test_client"):
            test_client = self.startTestClient(test_servers[0])

        with self.subTest("08_test_client_xds_config_exists"):
            self.assertXdsConfigExists(test_client)

        with self.subTest("09_test_server_received_rpcs_from_test_client"):
            self.assertSuccessfulRpcs(test_client)

        with self.subTest("10_round_robin"):
            num_rpcs = 100
            expected_rpcs_per_replica = num_rpcs / REPLICA_COUNT

            rpcs_by_peer = self.getClientRpcStats(
                test_client, num_rpcs
            ).rpcs_by_peer
            total_requests_received = sum(rpcs_by_peer[x] for x in rpcs_by_peer)
            self.assertEqual(
                total_requests_received, num_rpcs, "Wrong number of RPCS"
            )
            for server in test_servers:
                hostname = server.hostname
                self.assertIn(
                    hostname,
                    rpcs_by_peer,
                    f"Server {hostname} did not receive RPCs",
                )
                self.assertLessEqual(
                    abs(rpcs_by_peer[hostname] - expected_rpcs_per_replica),
                    1,
                    f"Wrong number of RPCs for server {hostname}",
                )


if __name__ == "__main__":
    absltest.main(failfast=True)



#!/usr/bin/env python3
# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper to upload Jenkins test results to BQ"""

from __future__ import print_function

import os
import sys
import time
import uuid

import six

gcp_utils_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "../../gcp/utils")
)
sys.path.append(gcp_utils_dir)
import big_query_utils

_DATASET_ID = "jenkins_test_results"
_DESCRIPTION = "Test results from master job run on Jenkins"
# 365 days in milliseconds
_EXPIRATION_MS = 365 * 24 * 60 * 60 * 1000
_PARTITION_TYPE = "DAY"
_PROJECT_ID = "grpc-testing"
_RESULTS_SCHEMA = [
    ("job_name", "STRING", "Name of Jenkins job"),
    ("build_id", "INTEGER", "Build ID of Jenkins job"),
    ("build_url", "STRING", "URL of Jenkins job"),
    ("test_name", "STRING", "Individual test name"),
    ("language", "STRING", "Language of test"),
    ("platform", "STRING", "Platform used for test"),
    ("config", "STRING", "Config used for test"),
    ("compiler", "STRING", "Compiler used for test"),
    ("iomgr_platform", "STRING", "Iomgr used for test"),
    ("result", "STRING", "Test result: PASSED, TIMEOUT, FAILED, or SKIPPED"),
    ("timestamp", "TIMESTAMP", "Timestamp of test run"),
    ("elapsed_time", "FLOAT", "How long test took to run"),
    ("cpu_estimated", "FLOAT", "Estimated CPU usage of test"),
    ("cpu_measured", "FLOAT", "Actual CPU usage of test"),
    ("return_code", "INTEGER", "Exit code of test"),
]
_INTEROP_RESULTS_SCHEMA = [
    ("job_name", "STRING", "Name of Jenkins/Kokoro job"),
    ("build_id", "INTEGER", "Build ID of Jenkins/Kokoro job"),
    ("build_url", "STRING", "URL of Jenkins/Kokoro job"),
    (
        "test_name",
        "STRING",
        "Unique test name combining client, server, and test_name",
    ),
    (
        "suite",
        "STRING",
        "Test suite: cloud_to_cloud, cloud_to_prod, or cloud_to_prod_auth",
    ),
    ("client", "STRING", "Client language"),
    ("server", "STRING", "Server host name"),
    ("test_case", "STRING", "Name of test case"),
    ("result", "STRING", "Test result: PASSED, TIMEOUT, FAILED, or SKIPPED"),
    ("timestamp", "TIMESTAMP", "Timestamp of test run"),
    ("elapsed_time", "FLOAT", "How long test took to run"),
]


def _get_build_metadata(test_results):
    """Add Kokoro build metadata to test_results based on environment
    variables set by Kokoro.
    """
    build_id = os.getenv("KOKORO_BUILD_NUMBER")
    build_url = (
        "https://source.cloud.google.com/results/invocations/%s"
        % os.getenv("KOKORO_BUILD_ID")
    )
    job_name = os.getenv("KOKORO_JOB_NAME")

    if build_id:
        test_results["build_id"] = build_id
    if build_url:
        test_results["build_url"] = build_url
    if job_name:
        test_results["job_name"] = job_name


def _insert_rows_with_retries(bq, bq_table, bq_rows):
    """Insert rows to bq table. Retry on error."""
    # BigQuery sometimes fails with large uploads, so batch 1,000 rows at a time.
    for i in range((len(bq_rows) // 1000) + 1):
        max_retries = 3
        for attempt in range(max_retries):
            if big_query_utils.insert_rows(
                bq,
                _PROJECT_ID,
                _DATASET_ID,
                bq_table,
                bq_rows[i * 1000 : (i + 1) * 1000],
            ):
                break
            else:
                if attempt < max_retries - 1:
                    print("Error uploading result to bigquery, will retry.")
                else:
                    print(
                        "Error uploading result to bigquery, all attempts"
                        " failed."
                    )
                    sys.exit(1)


def upload_results_to_bq(resultset, bq_table, extra_fields):
    """Upload test results to a BQ table.

    Args:
        resultset: dictionary generated by jobset.run
        bq_table: string name of table to create/upload results to in BQ
        extra_fields: dict with extra values that will be uploaded along with the results
    """
    bq = big_query_utils.create_big_query()
    big_query_utils.create_partitioned_table(
        bq,
        _PROJECT_ID,
        _DATASET_ID,
        bq_table,
        _RESULTS_SCHEMA,
        _DESCRIPTION,
        partition_type=_PARTITION_TYPE,
        expiration_ms=_EXPIRATION_MS,
    )

    bq_rows = []
    for shortname, results in six.iteritems(resultset):
        for result in results:
            test_results = {}
            _get_build_metadata(test_results)
            test_results["cpu_estimated"] = result.cpu_estimated
            test_results["cpu_measured"] = result.cpu_measured
            test_results["elapsed_time"] = "%.2f" % result.elapsed_time
            test_results["result"] = result.state
            test_results["return_code"] = result.returncode
            test_results["test_name"] = shortname
            test_results["timestamp"] = time.strftime("%Y-%m-%d %H:%M:%S")
            for field_name, field_value in six.iteritems(extra_fields):
                test_results[field_name] = field_value
            row = big_query_utils.make_row(str(uuid.uuid4()), test_results)
            bq_rows.append(row)
    _insert_rows_with_retries(bq, bq_table, bq_rows)


def upload_interop_results_to_bq(resultset, bq_table):
    """Upload interop test results to a BQ table.

    Args:
        resultset: dictionary generated by jobset.run
        bq_table: string name of table to create/upload results to in BQ
    """
    bq = big_query_utils.create_big_query()
    big_query_utils.create_partitioned_table(
        bq,
        _PROJECT_ID,
        _DATASET_ID,
        bq_table,
        _INTEROP_RESULTS_SCHEMA,
        _DESCRIPTION,
        partition_type=_PARTITION_TYPE,
        expiration_ms=_EXPIRATION_MS,
    )

    bq_rows = []
    for shortname, results in six.iteritems(resultset):
        for result in results:
            test_results = {}
            _get_build_metadata(test_results)
            test_results["elapsed_time"] = "%.2f" % result.elapsed_time
            test_results["result"] = result.state
            test_results["test_name"] = shortname
            test_results["suite"] = shortname.split(":")[0]
            test_results["client"] = shortname.split(":")[1]
            test_results["server"] = shortname.split(":")[2]
            test_results["test_case"] = shortname.split(":")[3]
            test_results["timestamp"] = time.strftime("%Y-%m-%d %H:%M:%S")
            row = big_query_utils.make_row(str(uuid.uuid4()), test_results)
            bq_rows.append(row)
    _insert_rows_with_retries(bq, bq_table, bq_rows)



# Copyright 2023 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

from absl.testing import absltest

from framework.test_app import server_app

# Alias
XdsTestServer = server_app.XdsTestServer

# Test values.
CANNED_IP: str = "10.0.0.43"
CANNED_RPC_PORT: int = 2222
CANNED_HOSTNAME: str = "test-server.local"
CANNED_XDS_HOST: str = "xds-test-server"
CANNED_XDS_PORT: int = 42


class ServerAppTest(absltest.TestCase):
    """Unit test for the XdsTestServer."""

    def test_constructor(self):
        xds_server = XdsTestServer(
            ip=CANNED_IP,
            rpc_port=CANNED_RPC_PORT,
            hostname=CANNED_HOSTNAME,
        )
        # Channels list empty.
        self.assertEmpty(xds_server.channels)

        # Test fields set as is.
        self.assertEqual(xds_server.ip, CANNED_IP)
        self.assertEqual(xds_server.rpc_port, CANNED_RPC_PORT)
        self.assertEqual(xds_server.hostname, CANNED_HOSTNAME)

        # Test optional argument defaults.
        self.assertEqual(xds_server.rpc_host, CANNED_IP)
        self.assertEqual(xds_server.maintenance_port, CANNED_RPC_PORT)
        self.assertEqual(xds_server.secure_mode, False)

    def test_xds_address(self):
        """Verifies the behavior of set_xds_address(), xds_address, xds_uri."""
        xds_server = XdsTestServer(
            ip=CANNED_IP,
            rpc_port=CANNED_RPC_PORT,
            hostname=CANNED_HOSTNAME,
        )
        self.assertEqual(xds_server.xds_uri, "", msg="Must be empty when unset")

        xds_server.set_xds_address(CANNED_XDS_HOST, CANNED_XDS_PORT)
        self.assertEqual(xds_server.xds_uri, "xds:///xds-test-server:42")

        xds_server.set_xds_address(CANNED_XDS_HOST, None)
        self.assertEqual(
            xds_server.xds_uri,
            "xds:///xds-test-server",
            msg="Must not contain ':port' when the port is not set",
        )

        xds_server.set_xds_address(None, None)
        self.assertEqual(xds_server.xds_uri, "", msg="Must be empty when reset")

        xds_server.set_xds_address(None, CANNED_XDS_PORT)
        self.assertEqual(
            xds_server.xds_uri,
            "",
            msg="Must be empty when only port is set",
        )


if __name__ == "__main__":
    absltest.main()



# Copyright 2021 The gRPC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import Tuple

from absl import flags
from absl.testing import absltest

from framework import xds_url_map_testcase
from framework.helpers import skips
from framework.test_app import client_app

# Type aliases
HostRule = xds_url_map_testcase.HostRule
PathMatcher = xds_url_map_testcase.PathMatcher
GcpResourceManager = xds_url_map_testcase.GcpResourceManager
DumpedXdsConfig = xds_url_map_testcase.DumpedXdsConfig
RpcTypeUnaryCall = xds_url_map_testcase.RpcTypeUnaryCall
RpcTypeEmptyCall = xds_url_map_testcase.RpcTypeEmptyCall
XdsTestClient = client_app.XdsTestClient
_Lang = skips.Lang

logger = logging.getLogger(__name__)
flags.adopt_module_key_flags(xds_url_map_testcase)

_NUM_RPCS = 150
_TEST_METADATA_KEY = "xds_md"
_TEST_METADATA_VALUE_UNARY = "unary_yranu"
_TEST_METADATA_VALUE_EMPTY = "empty_ytpme"
_TEST_METADATA_NUMERIC_KEY = "xds_md_numeric"
_TEST_METADATA_NUMERIC_VALUE = "159"

_TEST_METADATA = (
    (RpcTypeUnaryCall, _TEST_METADATA_KEY, _TEST_METADATA_VALUE_UNARY),
    (RpcTypeEmptyCall, _TEST_METADATA_KEY, _TEST_METADATA_VALUE_EMPTY),
    (
        RpcTypeUnaryCall,
        _TEST_METADATA_NUMERIC_KEY,
        _TEST_METADATA_NUMERIC_VALUE,
    ),
)


def _is_supported(config: skips.TestConfig) -> bool:
    if config.client_lang == _Lang.NODE:
        return config.version_gte("v1.3.x")
    return True


class TestExactMatch(xds_url_map_testcase.XdsUrlMapTestCase):
    @staticmethod
    def is_supported(config: skips.TestConfig) -> bool:
        return _is_supported(config)

    @staticmethod
    def url_map_change(
        host_rule: HostRule, path_matcher: PathMatcher
    ) -> Tuple[HostRule, PathMatcher]:
        path_matcher["routeRules"] = [
            {
                "priority": 0,
                # Header ExactMatch -> alternate_backend_service.
                # EmptyCall is sent with the metadata.
                "matchRules": [
                    {
                        "prefixMatch": "/",
                        "headerMatches": [
                            {
                                "headerName": _TEST_METADATA_KEY,
                                "exactMatch": _TEST_METADATA_VALUE_EMPTY,
                            }
                        ],
                    }
                ],
                "service": GcpResourceManager().alternative_backend_service(),
            }
        ]
        return host_rule, path_matcher

    def xds_config_validate(self, xds_config: DumpedXdsConfig):
        self.assertNumEndpoints(xds_config, 2)
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["name"],
            _TEST_METADATA_KEY,
        )
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["exactMatch"],
            _TEST_METADATA_VALUE_EMPTY,
        )

    def rpc_distribution_validate(self, test_client: XdsTestClient):
        rpc_distribution = self.configure_and_send(
            test_client,
            rpc_types=[RpcTypeEmptyCall],
            metadata=_TEST_METADATA,
            num_rpcs=_NUM_RPCS,
        )
        self.assertEqual(
            _NUM_RPCS, rpc_distribution.empty_call_alternative_service_rpc_count
        )


@absltest.skip("the xDS config is good, but distribution is wrong.")
class TestPrefixMatch(xds_url_map_testcase.XdsUrlMapTestCase):
    @staticmethod
    def is_supported(config: skips.TestConfig) -> bool:
        return _is_supported(config)

    @staticmethod
    def url_map_change(
        host_rule: HostRule, path_matcher: PathMatcher
    ) -> Tuple[HostRule, PathMatcher]:
        path_matcher["routeRules"] = [
            {
                "priority": 0,
                # Header PrefixMatch -> alternate_backend_service.
                # UnaryCall is sent with the metadata.
                "matchRules": [
                    {
                        "prefixMatch": "/",
                        "headerMatches": [
                            {
                                "headerName": _TEST_METADATA_KEY,
                                "prefixMatch": _TEST_METADATA_VALUE_UNARY[:2],
                            }
                        ],
                    }
                ],
                "service": GcpResourceManager().alternative_backend_service(),
            }
        ]
        return host_rule, path_matcher

    def xds_config_validate(self, xds_config: DumpedXdsConfig):
        self.assertNumEndpoints(xds_config, 2)
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["name"],
            _TEST_METADATA_KEY,
        )
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["prefixMatch"],
            _TEST_METADATA_VALUE_UNARY[:2],
        )

    def rpc_distribution_validate(self, test_client: XdsTestClient):
        rpc_distribution = self.configure_and_send(
            test_client,
            rpc_types=(RpcTypeUnaryCall,),
            metadata=_TEST_METADATA,
            num_rpcs=_NUM_RPCS,
        )
        self.assertEqual(
            _NUM_RPCS, rpc_distribution.unary_call_alternative_service_rpc_count
        )


class TestSuffixMatch(xds_url_map_testcase.XdsUrlMapTestCase):
    @staticmethod
    def is_supported(config: skips.TestConfig) -> bool:
        return _is_supported(config)

    @staticmethod
    def url_map_change(
        host_rule: HostRule, path_matcher: PathMatcher
    ) -> Tuple[HostRule, PathMatcher]:
        path_matcher["routeRules"] = [
            {
                "priority": 0,
                # Header SuffixMatch -> alternate_backend_service.
                # EmptyCall is sent with the metadata.
                "matchRules": [
                    {
                        "prefixMatch": "/",
                        "headerMatches": [
                            {
                                "headerName": _TEST_METADATA_KEY,
                                "suffixMatch": _TEST_METADATA_VALUE_EMPTY[-2:],
                            }
                        ],
                    }
                ],
                "service": GcpResourceManager().alternative_backend_service(),
            }
        ]
        return host_rule, path_matcher

    def xds_config_validate(self, xds_config: DumpedXdsConfig):
        self.assertNumEndpoints(xds_config, 2)
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["name"],
            _TEST_METADATA_KEY,
        )
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["suffixMatch"],
            _TEST_METADATA_VALUE_EMPTY[-2:],
        )

    def rpc_distribution_validate(self, test_client: XdsTestClient):
        rpc_distribution = self.configure_and_send(
            test_client,
            rpc_types=[RpcTypeEmptyCall],
            metadata=_TEST_METADATA,
            num_rpcs=_NUM_RPCS,
        )
        self.assertEqual(
            _NUM_RPCS, rpc_distribution.empty_call_alternative_service_rpc_count
        )


class TestPresentMatch(xds_url_map_testcase.XdsUrlMapTestCase):
    @staticmethod
    def is_supported(config: skips.TestConfig) -> bool:
        return _is_supported(config)

    @staticmethod
    def url_map_change(
        host_rule: HostRule, path_matcher: PathMatcher
    ) -> Tuple[HostRule, PathMatcher]:
        path_matcher["routeRules"] = [
            {
                "priority": 0,
                # Header 'xds_md_numeric' present -> alternate_backend_service.
                # UnaryCall is sent with the metadata, so will be sent to alternative.
                "matchRules": [
                    {
                        "prefixMatch": "/",
                        "headerMatches": [
                            {
                                "headerName": _TEST_METADATA_NUMERIC_KEY,
                                "presentMatch": True,
                            }
                        ],
                    }
                ],
                "service": GcpResourceManager().alternative_backend_service(),
            }
        ]
        return host_rule, path_matcher

    def xds_config_validate(self, xds_config: DumpedXdsConfig):
        self.assertNumEndpoints(xds_config, 2)
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["name"],
            _TEST_METADATA_NUMERIC_KEY,
        )
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["presentMatch"],
            True,
        )

    def rpc_distribution_validate(self, test_client: XdsTestClient):
        rpc_distribution = self.configure_and_send(
            test_client,
            rpc_types=(RpcTypeUnaryCall,),
            metadata=_TEST_METADATA,
            num_rpcs=_NUM_RPCS,
        )
        self.assertEqual(
            _NUM_RPCS, rpc_distribution.unary_call_alternative_service_rpc_count
        )


class TestInvertMatch(xds_url_map_testcase.XdsUrlMapTestCase):
    @staticmethod
    def is_supported(config: skips.TestConfig) -> bool:
        return _is_supported(config)

    @staticmethod
    def url_map_change(
        host_rule: HostRule, path_matcher: PathMatcher
    ) -> Tuple[HostRule, PathMatcher]:
        path_matcher["routeRules"] = [
            {
                "priority": 0,
                # Header invert ExactMatch -> alternate_backend_service.
                # UnaryCall is sent with the metadata, so will be sent to
                # default. EmptyCall will be sent to alternative.
                "matchRules": [
                    {
                        "prefixMatch": "/",
                        "headerMatches": [
                            {
                                "headerName": _TEST_METADATA_KEY,
                                "exactMatch": _TEST_METADATA_VALUE_UNARY,
                                "invertMatch": True,
                            }
                        ],
                    }
                ],
                "service": GcpResourceManager().alternative_backend_service(),
            }
        ]
        return host_rule, path_matcher

    def xds_config_validate(self, xds_config: DumpedXdsConfig):
        self.assertNumEndpoints(xds_config, 2)
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["name"],
            _TEST_METADATA_KEY,
        )
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["invertMatch"],
            True,
        )

    def rpc_distribution_validate(self, test_client: XdsTestClient):
        rpc_distribution = self.configure_and_send(
            test_client,
            rpc_types=[RpcTypeUnaryCall, RpcTypeEmptyCall],
            metadata=_TEST_METADATA,
            num_rpcs=_NUM_RPCS,
        )
        self.assertEqual(_NUM_RPCS, rpc_distribution.num_oks)
        self.assertEqual(
            0, rpc_distribution.unary_call_alternative_service_rpc_count
        )
        self.assertEqual(
            0, rpc_distribution.empty_call_default_service_rpc_count
        )


class TestRangeMatch(xds_url_map_testcase.XdsUrlMapTestCase):
    @staticmethod
    def is_supported(config: skips.TestConfig) -> bool:
        return _is_supported(config)

    @staticmethod
    def url_map_change(
        host_rule: HostRule, path_matcher: PathMatcher
    ) -> Tuple[HostRule, PathMatcher]:
        path_matcher["routeRules"] = [
            {
                "priority": 0,
                # Header 'xds_md_numeric' range [100,200] -> alternate_backend_service.
                # UnaryCall is sent with the metadata in range.
                "matchRules": [
                    {
                        "prefixMatch": "/",
                        "headerMatches": [
                            {
                                "headerName": _TEST_METADATA_NUMERIC_KEY,
                                "rangeMatch": {
                                    "rangeStart": "100",
                                    "rangeEnd": "200",
                                },
                            }
                        ],
                    }
                ],
                "service": GcpResourceManager().alternative_backend_service(),
            }
        ]
        return host_rule, path_matcher

    def xds_config_validate(self, xds_config: DumpedXdsConfig):
        self.assertNumEndpoints(xds_config, 2)
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["name"],
            _TEST_METADATA_NUMERIC_KEY,
        )
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["rangeMatch"]["start"],
            "100",
        )
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["rangeMatch"]["end"],
            "200",
        )

    def rpc_distribution_validate(self, test_client: XdsTestClient):
        rpc_distribution = self.configure_and_send(
            test_client,
            rpc_types=[RpcTypeUnaryCall, RpcTypeEmptyCall],
            metadata=_TEST_METADATA,
            num_rpcs=_NUM_RPCS,
        )
        self.assertEqual(_NUM_RPCS, rpc_distribution.num_oks)
        self.assertEqual(
            0, rpc_distribution.unary_call_default_service_rpc_count
        )
        self.assertEqual(
            0, rpc_distribution.empty_call_alternative_service_rpc_count
        )


class TestRegexMatch(xds_url_map_testcase.XdsUrlMapTestCase):
    @staticmethod
    def is_supported(config: skips.TestConfig) -> bool:
        return _is_supported(config)

    @staticmethod
    def url_map_change(
        host_rule: HostRule, path_matcher: PathMatcher
    ) -> Tuple[HostRule, PathMatcher]:
        path_matcher["routeRules"] = (
            [
                {
                    "priority": 0,
                    # Header RegexMatch -> alternate_backend_service.
                    # EmptyCall is sent with the metadata.
                    "matchRules": [
                        {
                            "prefixMatch": "/",
                            "headerMatches": [
                                {
                                    "headerName": _TEST_METADATA_KEY,
                                    "regexMatch": "^%s.*%s$"
                                    % (
                                        _TEST_METADATA_VALUE_EMPTY[:2],
                                        _TEST_METADATA_VALUE_EMPTY[-2:],
                                    ),
                                }
                            ],
                        }
                    ],
                    "service": GcpResourceManager().alternative_backend_service(),
                }
            ],
        )
        return host_rule, path_matcher

    def xds_config_validate(self, xds_config: DumpedXdsConfig):
        self.assertNumEndpoints(xds_config, 2)
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["name"],
            _TEST_METADATA_KEY,
        )
        self.assertEqual(
            xds_config.rds["virtualHosts"][0]["routes"][0]["match"]["headers"][
                0
            ]["safeRegexMatch"]["regex"],
            "^%s.*%s$"
            % (_TEST_METADATA_VALUE_EMPTY[:2], _TEST_METADATA_VALUE_EMPTY[-2:]),
        )

    def rpc_distribution_validate(self, test_client: XdsTestClient):
        rpc_distribution = self.configure_and_send(
            test_client,
            rpc_types=[RpcTypeEmptyCall],
            metadata=_TEST_METADATA,
            num_rpcs=_NUM_RPCS,
        )
        self.assertEqual(
            _NUM_RPCS, rpc_distribution.empty_call_alternative_service_rpc_count
        )


if __name__ == "__main__":
    absltest.main()



# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers to run docker instances as jobs."""

from __future__ import print_function

import json
import os
import subprocess
import sys
import tempfile
import time
import uuid

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
import jobset

_DEVNULL = open(os.devnull, "w")


def random_name(base_name):
    """Randomizes given base name."""
    return "%s_%s" % (base_name, uuid.uuid4())


def docker_kill(cid):
    """Kills a docker container. Returns True if successful."""
    return (
        subprocess.call(
            ["docker", "kill", str(cid)],
            stdin=subprocess.PIPE,
            stdout=_DEVNULL,
            stderr=subprocess.STDOUT,
        )
        == 0
    )


def docker_mapped_port(cid, port, timeout_seconds=15):
    """Get port mapped to internal given internal port for given container."""
    started = time.time()
    while time.time() - started < timeout_seconds:
        try:
            output = subprocess.check_output(
                "docker port %s %s" % (cid, port), stderr=_DEVNULL, shell=True
            ).decode()
            return int(output.split(":", 2)[1])
        except subprocess.CalledProcessError as e:
            pass
    raise Exception(
        "Failed to get exposed port %s for container %s." % (port, cid)
    )


def docker_ip_address(cid, timeout_seconds=15):
    """Get port mapped to internal given internal port for given container."""
    started = time.time()
    while time.time() - started < timeout_seconds:
        cmd = "docker inspect %s" % cid
        try:
            output = subprocess.check_output(
                cmd, stderr=_DEVNULL, shell=True
            ).decode()
            json_info = json.loads(output)
            assert len(json_info) == 1
            out = json_info[0]["NetworkSettings"]["IPAddress"]
            if not out:
                continue
            return out
        except subprocess.CalledProcessError as e:
            pass
    raise Exception(
        "Non-retryable error: Failed to get ip address of container %s." % cid
    )


def wait_for_healthy(cid, shortname, timeout_seconds):
    """Wait timeout_seconds for the container to become healthy"""
    started = time.time()
    while time.time() - started < timeout_seconds:
        try:
            output = subprocess.check_output(
                [
                    "docker",
                    "inspect",
                    '--format="{{.State.Health.Status}}"',
                    cid,
                ],
                stderr=_DEVNULL,
            ).decode()
            if output.strip("\n") == "healthy":
                return
        except subprocess.CalledProcessError as e:
            pass
        time.sleep(1)
    raise Exception(
        "Timed out waiting for %s (%s) to pass health check" % (shortname, cid)
    )


def finish_jobs(jobs, suppress_failure=True):
    """Kills given docker containers and waits for corresponding jobs to finish"""
    for job in jobs:
        job.kill(suppress_failure=suppress_failure)

    while any(job.is_running() for job in jobs):
        time.sleep(1)


def image_exists(image):
    """Returns True if given docker image exists."""
    return (
        subprocess.call(
            ["docker", "inspect", image],
            stdin=subprocess.PIPE,
            stdout=_DEVNULL,
            stderr=subprocess.STDOUT,
        )
        == 0
    )


def remove_image(image, skip_nonexistent=False, max_retries=10):
    """Attempts to remove docker image with retries."""
    if skip_nonexistent and not image_exists(image):
        return True
    for attempt in range(0, max_retries):
        if (
            subprocess.call(
                ["docker", "rmi", "-f", image],
                stdin=subprocess.PIPE,
                stdout=_DEVNULL,
                stderr=subprocess.STDOUT,
            )
            == 0
        ):
            return True
        time.sleep(2)
    print("Failed to remove docker image %s" % image)
    return False


class DockerJob:
    """Encapsulates a job"""

    def __init__(self, spec):
        self._spec = spec
        self._job = jobset.Job(
            spec, newline_on_success=True, travis=True, add_env={}
        )
        self._container_name = spec.container_name

    def mapped_port(self, port):
        return docker_mapped_port(self._container_name, port)

    def ip_address(self):
        return docker_ip_address(self._container_name)

    def wait_for_healthy(self, timeout_seconds):
        wait_for_healthy(
            self._container_name, self._spec.shortname, timeout_seconds
        )

    def kill(self, suppress_failure=False):
        """Sends kill signal to the container."""
        if suppress_failure:
            self._job.suppress_failure_message()
        return docker_kill(self._container_name)

    def is_running(self):
        """Polls a job and returns True if given job is still running."""
        return self._job.state() == jobset._RUNNING



#!/usr/bin/env python3
#
# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Runs the entire bm_*.py pipeline, and possible comments on the PR """

import argparse
import multiprocessing
import os
import random
import subprocess
import sys

sys.path.append(
    os.path.join(
        os.path.dirname(sys.argv[0]), "..", "..", "run_tests", "python_utils"
    )
)

sys.path.append(
    os.path.join(
        os.path.dirname(sys.argv[0]),
        "..",
        "..",
        "..",
        "run_tests",
        "python_utils",
    )
)

import bm_build
import bm_constants
import bm_diff
import bm_run
import check_on_pr
import jobset


def _args():
    argp = argparse.ArgumentParser(
        description="Perform diff on microbenchmarks"
    )
    argp.add_argument(
        "-t",
        "--track",
        choices=sorted(bm_constants._INTERESTING),
        nargs="+",
        default=sorted(bm_constants._INTERESTING),
        help="Which metrics to track",
    )
    argp.add_argument(
        "-b",
        "--benchmarks",
        nargs="+",
        choices=bm_constants._AVAILABLE_BENCHMARK_TESTS,
        default=bm_constants._AVAILABLE_BENCHMARK_TESTS,
        help="Which benchmarks to run",
    )
    argp.add_argument(
        "-d",
        "--diff_base",
        type=str,
        help="Commit or branch to compare the current one to",
    )
    argp.add_argument(
        "-o",
        "--old",
        default="old",
        type=str,
        help='Name of baseline run to compare to. Usually just called "old"',
    )
    argp.add_argument(
        "-r",
        "--regex",
        type=str,
        default="",
        help="Regex to filter benchmarks run",
    )
    argp.add_argument(
        "-l",
        "--loops",
        type=int,
        default=10,
        help=(
            "Number of times to loops the benchmarks. More loops cuts down on"
            " noise"
        ),
    )
    argp.add_argument(
        "-j",
        "--jobs",
        type=int,
        default=multiprocessing.cpu_count(),
        help="Number of CPUs to use",
    )
    argp.add_argument(
        "--pr_comment_name",
        type=str,
        default="microbenchmarks",
        help="Name that Jenkins will use to comment on the PR",
    )
    args = argp.parse_args()
    assert args.diff_base or args.old, "One of diff_base or old must be set!"
    if args.loops < 3:
        print("WARNING: This run will likely be noisy. Increase loops.")
    return args


def eintr_be_gone(fn):
    """Run fn until it doesn't stop because of EINTR"""

    def inner(*args):
        while True:
            try:
                return fn(*args)
            except IOError as e:
                if e.errno != errno.EINTR:
                    raise

    return inner


def main(args):
    bm_build.build("new", args.benchmarks, args.jobs)

    old = args.old
    if args.diff_base:
        old = "old"
        where_am_i = subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"]
        ).strip()
        subprocess.check_call(["git", "checkout", args.diff_base])
        try:
            bm_build.build(old, args.benchmarks, args.jobs)
        finally:
            subprocess.check_call(["git", "checkout", where_am_i])
            subprocess.check_call(["git", "submodule", "update"])

    jobs_list = []
    jobs_list += bm_run.create_jobs(
        "new", args.benchmarks, args.loops, args.regex
    )
    jobs_list += bm_run.create_jobs(
        old, args.benchmarks, args.loops, args.regex
    )

    # shuffle all jobs to eliminate noise from GCE CPU drift
    random.shuffle(jobs_list, random.SystemRandom().random)
    jobset.run(jobs_list, maxjobs=args.jobs)

    diff, note, significance = bm_diff.diff(
        args.benchmarks, args.loops, args.regex, args.track, old, "new"
    )
    if diff:
        text = "[%s] Performance differences noted:\n%s" % (
            args.pr_comment_name,
            diff,
        )
    else:
        text = (
            "[%s] No significant performance differences" % args.pr_comment_name
        )
    if note:
        text = note + "\n\n" + text
    print("%s" % text)
    check_on_pr.check_on_pr("Benchmark", "```\n%s\n```" % text)


if __name__ == "__main__":
    args = _args()
    main(args)



""" Multicast DNS Service Discovery for Python, v0.14-wmcbrine
    Copyright 2003 Paul Scott-Murphy, 2014 William McBrine

    This module provides a framework for the use of DNS Service Discovery
    using IP multicast.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
    USA
"""

from functools import lru_cache
from typing import Set

from .._exceptions import BadTypeInNameException
from ..const import (
    _HAS_A_TO_Z,
    _HAS_ASCII_CONTROL_CHARS,
    _HAS_ONLY_A_TO_Z_NUM_HYPHEN,
    _HAS_ONLY_A_TO_Z_NUM_HYPHEN_UNDERSCORE,
    _LOCAL_TRAILER,
    _NONTCP_PROTOCOL_LOCAL_TRAILER,
    _TCP_PROTOCOL_LOCAL_TRAILER,
)


@lru_cache(maxsize=512)
def service_type_name(type_: str, *, strict: bool = True) -> str:  # pylint: disable=too-many-branches
    """
    Validate a fully qualified service name, instance or subtype. [rfc6763]

    Returns fully qualified service name.

    Domain names used by mDNS-SD take the following forms:

                   <sn> . <_tcp|_udp> . local.
      <Instance> . <sn> . <_tcp|_udp> . local.
      <sub>._sub . <sn> . <_tcp|_udp> . local.

    1) must end with 'local.'

      This is true because we are implementing mDNS and since the 'm' means
      multi-cast, the 'local.' domain is mandatory.

    2) local is preceded with either '_udp.' or '_tcp.' unless
       strict is False

    3) service name <sn> precedes <_tcp|_udp> unless
       strict is False

      The rules for Service Names [RFC6335] state that they may be no more
      than fifteen characters long (not counting the mandatory underscore),
      consisting of only letters, digits, and hyphens, must begin and end
      with a letter or digit, must not contain consecutive hyphens, and
      must contain at least one letter.

    The instance name <Instance> and sub type <sub> may be up to 63 bytes.

    The portion of the Service Instance Name is a user-
    friendly name consisting of arbitrary Net-Unicode text [RFC5198]. It
    MUST NOT contain ASCII control characters (byte values 0x00-0x1F and
    0x7F) [RFC20] but otherwise is allowed to contain any characters,
    without restriction, including spaces, uppercase, lowercase,
    punctuation -- including dots -- accented characters, non-Roman text,
    and anything else that may be represented using Net-Unicode.

    :param type_: Type, SubType or service name to validate
    :return: fully qualified service name (eg: _http._tcp.local.)
    """
    if len(type_) > 256:
        # https://datatracker.ietf.org/doc/html/rfc6763#section-7.2
        raise BadTypeInNameException("Full name (%s) must be > 256 bytes" % type_)

    if type_.endswith((_TCP_PROTOCOL_LOCAL_TRAILER, _NONTCP_PROTOCOL_LOCAL_TRAILER)):
        remaining = type_[: -len(_TCP_PROTOCOL_LOCAL_TRAILER)].split('.')
        trailer = type_[-len(_TCP_PROTOCOL_LOCAL_TRAILER) :]
        has_protocol = True
    elif strict:
        raise BadTypeInNameException(
            "Type '%s' must end with '%s' or '%s'"
            % (type_, _TCP_PROTOCOL_LOCAL_TRAILER, _NONTCP_PROTOCOL_LOCAL_TRAILER)
        )
    elif type_.endswith(_LOCAL_TRAILER):
        remaining = type_[: -len(_LOCAL_TRAILER)].split('.')
        trailer = type_[-len(_LOCAL_TRAILER) + 1 :]
        has_protocol = False
    else:
        raise BadTypeInNameException(f"Type '{type_}' must end with '{_LOCAL_TRAILER}'")

    if strict or has_protocol:
        service_name = remaining.pop()
        if not service_name:
            raise BadTypeInNameException("No Service name found")

        if len(remaining) == 1 and len(remaining[0]) == 0:
            raise BadTypeInNameException("Type '%s' must not start with '.'" % type_)

        if service_name[0] != '_':
            raise BadTypeInNameException("Service name (%s) must start with '_'" % service_name)

        test_service_name = service_name[1:]

        if strict and len(test_service_name) > 15:
            # https://datatracker.ietf.org/doc/html/rfc6763#section-7.2
            raise BadTypeInNameException("Service name (%s) must be <= 15 bytes" % test_service_name)

        if '--' in test_service_name:
            raise BadTypeInNameException("Service name (%s) must not contain '--'" % test_service_name)

        if '-' in (test_service_name[0], test_service_name[-1]):
            raise BadTypeInNameException(
                "Service name (%s) may not start or end with '-'" % test_service_name
            )

        if not _HAS_A_TO_Z.search(test_service_name):
            raise BadTypeInNameException(
                "Service name (%s) must contain at least one letter (eg: 'A-Z')" % test_service_name
            )

        allowed_characters_re = (
            _HAS_ONLY_A_TO_Z_NUM_HYPHEN if strict else _HAS_ONLY_A_TO_Z_NUM_HYPHEN_UNDERSCORE
        )

        if not allowed_characters_re.search(test_service_name):
            raise BadTypeInNameException(
                "Service name (%s) must contain only these characters: "
                "A-Z, a-z, 0-9, hyphen ('-')%s" % (test_service_name, "" if strict else ", underscore ('_')")
            )
    else:
        service_name = ''

    if remaining and remaining[-1] == '_sub':
        remaining.pop()
        if len(remaining) == 0 or len(remaining[0]) == 0:
            raise BadTypeInNameException("_sub requires a subtype name")

    if len(remaining) > 1:
        remaining = ['.'.join(remaining)]

    if remaining:
        length = len(remaining[0].encode('utf-8'))
        if length > 63:
            raise BadTypeInNameException("Too long: '%s'" % remaining[0])

        if _HAS_ASCII_CONTROL_CHARS.search(remaining[0]):
            raise BadTypeInNameException(
                "Ascii control character 0x00-0x1F and 0x7F illegal in '%s'" % remaining[0]
            )

    return service_name + trailer


def possible_types(name: str) -> Set[str]:
    """Build a set of all possible types from a fully qualified name."""
    labels = name.split('.')
    label_count = len(labels)
    types = set()
    for count in range(label_count):
        parts = labels[label_count - count - 4 :]
        if not parts[0].startswith('_'):
            break
        types.add('.'.join(parts))
    return types


cached_possible_types = lru_cache(maxsize=256)(possible_types)




import cython

from .._cache cimport DNSCache
from .._dns cimport DNSEntry, DNSPointer, DNSQuestion, DNSRecord
from .incoming cimport DNSIncoming


cdef cython.uint _CLASS_UNIQUE
cdef cython.uint _DNS_PACKET_HEADER_LEN
cdef cython.uint _FLAGS_QR_MASK
cdef cython.uint _FLAGS_QR_QUERY
cdef cython.uint _FLAGS_QR_RESPONSE
cdef cython.uint _FLAGS_TC
cdef cython.uint _MAX_MSG_ABSOLUTE
cdef cython.uint _MAX_MSG_TYPICAL

cdef object TYPE_CHECKING

cdef object PACK_BYTE
cdef object PACK_SHORT
cdef object PACK_LONG

cdef object STATE_INIT
cdef object STATE_FINISHED

cdef object LOGGING_IS_ENABLED_FOR
cdef object LOGGING_DEBUG

cdef cython.tuple BYTE_TABLE

cdef class DNSOutgoing:

    cdef public unsigned int flags
    cdef public bint finished
    cdef public object id
    cdef public bint multicast
    cdef public cython.list packets_data
    cdef public cython.dict names
    cdef public cython.list data
    cdef public unsigned int size
    cdef public bint allow_long
    cdef public object state
    cdef public cython.list questions
    cdef public cython.list answers
    cdef public cython.list authorities
    cdef public cython.list additionals

    cdef _reset_for_next_packet(self)

    cdef _write_byte(self, object value)

    cdef _insert_short_at_start(self, object value)

    cdef _replace_short(self, object index, object value)

    cdef _write_int(self, object value)

    cdef cython.bint _write_question(self, DNSQuestion question)

    @cython.locals(
        d=cython.bytes,
        data_view=cython.list,
        length=cython.uint
    )
    cdef cython.bint _write_record(self, DNSRecord record, object now)

    cdef _write_record_class(self, DNSEntry record)

    @cython.locals(
        start_size_int=object
    )
    cdef cython.bint _check_data_limit_or_rollback(self, cython.uint start_data_length, cython.uint start_size)

    cdef _write_questions_from_offset(self, object questions_offset)

    cdef _write_answers_from_offset(self, object answer_offset)

    cdef _write_records_from_offset(self, cython.list records, object offset)

    cdef _has_more_to_add(self, object questions_offset, object answer_offset, object authority_offset, object additional_offset)

    cdef _write_ttl(self, DNSRecord record, object now)

    @cython.locals(
        labels=cython.list,
        label=cython.str,
        index=cython.uint,
        start_size=cython.uint,
        name_length=cython.uint,
    )
    cpdef write_name(self, cython.str name)

    cpdef write_short(self, object value)

    cpdef write_string(self, cython.bytes value)

    cpdef _write_utf(self, cython.str value)

    @cython.locals(
        questions_offset=object,
        answer_offset=object,
        authority_offset=object,
        additional_offset=object,
        questions_written=object,
        answers_written=object,
        authorities_written=object,
        additionals_written=object,
    )
    cdef _packets(self)

    cpdef add_question_or_all_cache(self, DNSCache cache, object now, str name, object type_, object class_)

    cpdef add_question_or_one_cache(self, DNSCache cache, object now, str name, object type_, object class_)

    cpdef add_question(self, DNSQuestion question)

    cpdef add_answer(self, DNSIncoming inp, DNSRecord record)

    @cython.locals(now_float=cython.float)
    cpdef add_answer_at_time(self, DNSRecord record, object now)

    cpdef add_authorative_answer(self, DNSPointer record)

    cpdef add_additional_answer(self, DNSRecord record)

    cpdef is_query(self)

    cpdef is_response(self)



# Copyright 2023 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common functionality for bin/ python helpers."""
import atexit
import signal
import sys

from absl import logging

from framework import xds_flags
from framework import xds_k8s_flags
from framework.infrastructure import gcp
from framework.infrastructure import k8s
from framework.test_app import client_app
from framework.test_app import server_app
from framework.test_app.runners.k8s import gamma_server_runner
from framework.test_app.runners.k8s import k8s_xds_client_runner
from framework.test_app.runners.k8s import k8s_xds_server_runner

logger = logging.get_absl_logger()

# Type aliases
KubernetesClientRunner = k8s_xds_client_runner.KubernetesClientRunner
KubernetesServerRunner = k8s_xds_server_runner.KubernetesServerRunner
GammaServerRunner = gamma_server_runner.GammaServerRunner
_XdsTestServer = server_app.XdsTestServer
_XdsTestClient = client_app.XdsTestClient


def make_client_namespace(
    k8s_api_manager: k8s.KubernetesApiManager,
    namespace_name: str = None,
) -> k8s.KubernetesNamespace:
    if not namespace_name:
        namespace_name: str = KubernetesClientRunner.make_namespace_name(
            xds_flags.RESOURCE_PREFIX.value, xds_flags.RESOURCE_SUFFIX.value
        )
    return k8s.KubernetesNamespace(k8s_api_manager, namespace_name)


def make_client_runner(
    namespace: k8s.KubernetesNamespace,
    gcp_api_manager: gcp.api.GcpApiManager,
    port_forwarding: bool = False,
    reuse_namespace: bool = True,
    mode: str = "default",
) -> KubernetesClientRunner:
    # KubernetesClientRunner arguments.
    runner_kwargs = dict(
        deployment_name=xds_flags.CLIENT_NAME.value,
        image_name=xds_k8s_flags.CLIENT_IMAGE.value,
        td_bootstrap_image=xds_k8s_flags.TD_BOOTSTRAP_IMAGE.value,
        gcp_project=xds_flags.PROJECT.value,
        gcp_api_manager=gcp_api_manager,
        gcp_service_account=xds_k8s_flags.GCP_SERVICE_ACCOUNT.value,
        xds_server_uri=xds_flags.XDS_SERVER_URI.value,
        network=xds_flags.NETWORK.value,
        stats_port=xds_flags.CLIENT_PORT.value,
        reuse_namespace=reuse_namespace,
        debug_use_port_forwarding=port_forwarding,
    )

    if mode == "secure":
        runner_kwargs.update(
            deployment_template="client-secure.deployment.yaml"
        )
    return KubernetesClientRunner(namespace, **runner_kwargs)


def make_server_namespace(
    k8s_api_manager: k8s.KubernetesApiManager,
    server_runner: KubernetesServerRunner = KubernetesServerRunner,
) -> k8s.KubernetesNamespace:
    namespace_name: str = server_runner.make_namespace_name(
        xds_flags.RESOURCE_PREFIX.value, xds_flags.RESOURCE_SUFFIX.value
    )
    return k8s.KubernetesNamespace(k8s_api_manager, namespace_name)


def make_server_runner(
    namespace: k8s.KubernetesNamespace,
    gcp_api_manager: gcp.api.GcpApiManager,
    port_forwarding: bool = False,
    reuse_namespace: bool = True,
    reuse_service: bool = False,
    mode: str = "default",
) -> KubernetesServerRunner:
    # KubernetesServerRunner arguments.
    runner_kwargs = dict(
        deployment_name=xds_flags.SERVER_NAME.value,
        image_name=xds_k8s_flags.SERVER_IMAGE.value,
        td_bootstrap_image=xds_k8s_flags.TD_BOOTSTRAP_IMAGE.value,
        xds_server_uri=xds_flags.XDS_SERVER_URI.value,
        gcp_project=xds_flags.PROJECT.value,
        gcp_api_manager=gcp_api_manager,
        gcp_service_account=xds_k8s_flags.GCP_SERVICE_ACCOUNT.value,
        network=xds_flags.NETWORK.value,
        reuse_namespace=reuse_namespace,
        reuse_service=reuse_service,
        debug_use_port_forwarding=port_forwarding,
    )

    server_runner = KubernetesServerRunner
    if mode == "secure":
        runner_kwargs["deployment_template"] = "server-secure.deployment.yaml"
    elif mode == "gamma":
        server_runner = GammaServerRunner

    return server_runner(namespace, **runner_kwargs)


def _ensure_atexit(signum, frame):
    """Needed to handle signals or atexit handler won't be called."""
    del frame

    # Pylint is wrong about "Module 'signal' has no 'Signals' member":
    # https://docs.python.org/3/library/signal.html#signal.Signals
    sig = signal.Signals(signum)  # pylint: disable=no-member
    logger.warning("Caught %r, initiating graceful shutdown...\n", sig)
    sys.exit(1)


def _graceful_exit(
    server_runner: KubernetesServerRunner, client_runner: KubernetesClientRunner
):
    """Stop port forwarding processes."""
    client_runner.stop_pod_dependencies()
    server_runner.stop_pod_dependencies()


def register_graceful_exit(
    server_runner: KubernetesServerRunner, client_runner: KubernetesClientRunner
):
    atexit.register(_graceful_exit, server_runner, client_runner)
    for signum in (signal.SIGTERM, signal.SIGHUP, signal.SIGINT):
        signal.signal(signum, _ensure_atexit)


def get_client_pod(
    client_runner: KubernetesClientRunner, deployment_name: str
) -> k8s.V1Pod:
    client_deployment: k8s.V1Deployment
    client_deployment = client_runner.k8s_namespace.get_deployment(
        deployment_name
    )
    client_pod_name: str = client_runner._wait_deployment_pod_count(
        client_deployment
    )[0]
    return client_runner._wait_pod_started(client_pod_name)


def get_server_pod(
    server_runner: KubernetesServerRunner, deployment_name: str
) -> k8s.V1Pod:
    server_deployment: k8s.V1Deployment
    server_deployment = server_runner.k8s_namespace.get_deployment(
        deployment_name
    )
    server_pod_name: str = server_runner._wait_deployment_pod_count(
        server_deployment
    )[0]
    return server_runner._wait_pod_started(server_pod_name)


def get_test_server_for_pod(
    server_runner: KubernetesServerRunner, server_pod: k8s.V1Pod, **kwargs
) -> _XdsTestServer:
    return server_runner._xds_test_server_for_pod(server_pod, **kwargs)


def get_test_client_for_pod(
    client_runner: KubernetesClientRunner, client_pod: k8s.V1Pod, **kwargs
) -> _XdsTestClient:
    return client_runner._xds_test_client_for_pod(client_pod, **kwargs)



#!/usr/bin/env python
# Copyright 2019 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

"""
A chain with multiple intermediates with different subjectKeyIdentifiers and
notBefore dates, for testing path bulding prioritization.
"""

import sys
sys.path += ['../..']

import gencerts

DATE_A = '150101120000Z'
DATE_B = '150102120000Z'
DATE_C = '150103120000Z'
DATE_Z = '180101120000Z'


root = gencerts.create_self_signed_root_certificate('Root')
root.set_validity_range(DATE_A, DATE_Z)

int_matching_ski_a = gencerts.create_intermediate_certificate('Intermediate',
                                                              root)
int_matching_ski_a.set_validity_range(DATE_A, DATE_Z)

int_matching_ski_b = gencerts.create_intermediate_certificate('Intermediate',
                                                              root)
int_matching_ski_b.set_validity_range(DATE_B, DATE_Z)
int_matching_ski_b.set_key(int_matching_ski_a.get_key())

int_matching_ski_c = gencerts.create_intermediate_certificate('Intermediate',
                                                              root)
int_matching_ski_c.set_validity_range(DATE_C, DATE_Z)
int_matching_ski_c.set_key(int_matching_ski_a.get_key())

# For some reason, OpenSSL seems to require disabling SKID and AKID on the
# parent cert in order to generate an intermediate cert without a SKID.
root2 = gencerts.create_self_signed_root_certificate('Root')
root2.set_key(root.get_key())
section = root2.config.get_section('signing_ca_ext')
section.remove_property('subjectKeyIdentifier')
section.remove_property('authorityKeyIdentifier')

int_no_ski_a = gencerts.create_intermediate_certificate('Intermediate', root2)
int_no_ski_a.set_validity_range(DATE_A, DATE_Z)
int_no_ski_a.set_key(int_matching_ski_a.get_key())
section = int_no_ski_a.config.get_section('req_ext')
section.remove_property('subjectKeyIdentifier')

int_no_ski_b = gencerts.create_intermediate_certificate('Intermediate', root2)
int_no_ski_b.set_validity_range(DATE_B, DATE_Z)
int_no_ski_b.set_key(int_matching_ski_a.get_key())
section = int_no_ski_b.config.get_section('req_ext')
section.remove_property('subjectKeyIdentifier')

int_no_ski_c = gencerts.create_intermediate_certificate('Intermediate', root2)
int_no_ski_c.set_validity_range(DATE_C, DATE_Z)
int_no_ski_c.set_key(int_matching_ski_a.get_key())
section = int_no_ski_c.config.get_section('req_ext')
section.remove_property('subjectKeyIdentifier')

int_different_ski_a = gencerts.create_intermediate_certificate('Intermediate',
                                                              root)
int_different_ski_a.set_validity_range(DATE_A, DATE_Z)

int_different_ski_b = gencerts.create_intermediate_certificate('Intermediate',
                                                              root)
int_different_ski_b.set_validity_range(DATE_B, DATE_Z)
int_different_ski_b.set_key(int_different_ski_a.get_key())

int_different_ski_c = gencerts.create_intermediate_certificate('Intermediate',
                                                              root)
int_different_ski_c.set_validity_range(DATE_C, DATE_Z)
int_different_ski_c.set_key(int_different_ski_a.get_key())

target = gencerts.create_end_entity_certificate('Target', int_matching_ski_a)
target.set_validity_range(DATE_A, DATE_Z)


gencerts.write_chain('The root', [root], out_pem='root.pem')

gencerts.write_chain(
    'Intermediate with matching subjectKeyIdentifier and notBefore A',
    [int_matching_ski_a], out_pem='int_matching_ski_a.pem')

gencerts.write_chain(
    'Intermediate with matching subjectKeyIdentifier and notBefore B',
    [int_matching_ski_b], out_pem='int_matching_ski_b.pem')

gencerts.write_chain(
    'Intermediate with matching subjectKeyIdentifier and notBefore C',
    [int_matching_ski_c], out_pem='int_matching_ski_c.pem')

gencerts.write_chain(
    'Intermediate with no subjectKeyIdentifier and notBefore A',
    [int_no_ski_a], out_pem='int_no_ski_a.pem')

gencerts.write_chain(
    'Intermediate with no subjectKeyIdentifier and notBefore B',
    [int_no_ski_b], out_pem='int_no_ski_b.pem')

gencerts.write_chain(
    'Intermediate with no subjectKeyIdentifier and notBefore C',
    [int_no_ski_c], out_pem='int_no_ski_c.pem')

gencerts.write_chain(
    'Intermediate with different subjectKeyIdentifier and notBefore A',
    [int_different_ski_a], out_pem='int_different_ski_a.pem')

gencerts.write_chain(
    'Intermediate with different subjectKeyIdentifier and notBefore B',
    [int_different_ski_b], out_pem='int_different_ski_b.pem')

gencerts.write_chain(
    'Intermediate with different subjectKeyIdentifier and notBefore C',
    [int_different_ski_c], out_pem='int_different_ski_c.pem')

gencerts.write_chain('The target', [target], out_pem='target.pem')

